#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
SPECEX - SPECtra EXtractor.

This module provides utility functions to handle spectroscopic datacubes.

Copyright (C) 2022-2024  Maurizio D'Addona <mauritiusdadd@gmail.com>
"""
from __future__ import annotations

import os
import re
import sys
import shutil
import logging
import argparse
import warnings

from numbers import Number
from typing import Optional, Union, Tuple, List, Dict, Callable, Any

import numpy as np
from scipy.ndimage import gaussian_filter, gaussian_filter1d

from astropy import units
from astropy.io import fits
from astropy.wcs import WCS
from astropy.coordinates import SkyCoord
from astropy.nddata.utils import Cutout2D

from specex.utils import get_pc_transform_params, rotate_data, get_pbar
from specex.utils import HAS_REGION, parse_regionfile, simple_pbar_callback
from specex.exceptions import get_ipython_embedder

if HAS_REGION:
    import regions

KNOWN_SPEC_EXT_NAMES = ['spec', 'spectrum', 'flux', 'data', 'sci', 'science']
KNOWN_VARIANCE_EXT_NAMES = ['stat', 'stats', 'var', 'variance', 'noise', 'err']
KNOWN_INVAR_EXT_NAMES = ['ivar', 'ivariance']
KNOWN_MASK_EXT_NAMES = ['mask', 'platemask', 'footprint', 'dq', 'nan_mask']
KNOWN_WAVE_EXT_NAMES = ['wave', 'wavelenght', 'lambda', 'lam']
KNOWN_RCURVE_EXT_NAMES = ['r', 'reso', 'resolution', 'rcurve', 'wd']
KNOWN_RGB_EXT_NAMES = ['r', 'g', 'b', 'red', 'green', 'blue']


class SpectralCube:
    """Class to handle datacubes."""

    def __init__(self) -> None:
        self.filename: Union[str, None] = None
        self.hdul: Union[fits.HDUList, None] = None
        self.spec_hdu: Union[fits.ImageHDU, None] = None
        self.var_hdu: Union[fits.ImageHDU, None] = None
        self.var_hdu: Union[fits.ImageHDU, None] = None
        self.mask_hdu: Union[fits.ImageHDU, None] = None
        self.wd_hdu: Union[fits.ImageHDU, None] = None
        self.spec_wcs: Union[WCS, None] = None
        self.var_wcs: Union[WCS, None] = None
        self.wd_wcs: Union[WCS, None] = None

    def __del__(self) -> None:
        """
        Do cleanup when this object is deleted.

        Returns
        -------
        None.

        """
        self.close()

    def __enter__(self) -> SpectralCube:
        """
        Enter the context manager.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return self

    def __exit__(self, *args) -> None:
        """
        Exit the contex manager.

        Parameters
        ----------
        *args : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        self.close()

    def getBaseName(self):
        bname = os.path.basename(self.filename)
        return os.path.splitext(bname)[0]

    def close(self) -> None:
        """
        Clean up on closing.

        Returns
        -------
        None.

        """
        if self.hdul is not None:
            self.hdul.close()

    def getSpecWCS(self) -> WCS:
        """
        Get the WCS of the spectrum datacube.

        Returns
        -------
        wcs
            The WCS of the spectral HDU.

        """
        if self.spec_wcs is None:
            self.spec_wcs = WCS(self.spec_hdu.header)
        return self.spec_wcs

    def getVarWCS(self) -> WCS:
        """
        Get the WCS of the variance datacube.

        Returns
        -------
        wcs
            The WCS of the variance HDU.
        """
        if self.var_wcs is None:
            self.var_wcs = WCS(self.var_hdu.header)
            if not (self.var_wcs.has_celestial and self.var_wcs.has_spectral):
                return self.getSpecWCS()
        return self.var_wcs

    def getWdWCS(self) -> WCS:
        """
        Get the WCS of the variance datacube.

        Returns
        -------
        wcs
            The WCS of the resolution curve HDU.
        """
        if self.wd_wcs is None:
            self.wd_wcs = WCS(self.wd_hdu.header)
            if not (self.wd_wcs.has_celestial and self.wd_wcs.has_spectral):
                return self.getSpecWCS()
        return self.wd_wcs

    def getSpatialSizePixels(
        self
    ) -> List[Union[None, int]]:
        """
        Return spatial dimensions in pixels.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        if self.spec_hdu is None:
            return [None, None]
        else:
            return self.spec_hdu.data.shape[1:]

    @classmethod
    def open(cls, filename: str,
             spec_hdu_index: Optional[Union[int, str]] = None,
             var_hdu_index: Optional[Union[int, str]] = None,
             mask_hdu_index: Optional[Union[int, str]] = None,
             wd_hdu_index: Optional[Union[int, str]] = None,
             mode: Optional[str] = None) -> SpectralCube:
        """
        Open a datacube from a FITS file.

        Parameters
        ----------
        filename : str
            The file path.
        spec_hdu_index : Optional[Union[int, str]], optional
            The index or name of the HDU containing the spectrum.
            The default is None.
        var_hdu_index : Optional[Union[int, str]], optional
            The index or name of the HDU containing the variance.
            The default is None.
        mask_hdu_index : Optional[Union[int, str]], optional
            The index or name of the HDU containing the nans maks.
            The default is None.
        wd_hdu_index : Optional[Union[int, str]], optional
            The index or name of the HDU containing the R matrix.
            The default is None.
        mode : Optional[str], optional
            The mode passed to astropy.io.fits.open

        Returns
        -------
        SpectralCube.

        """
        new_cube = cls()
        new_cube.filename = filename
        new_cube.hdul = fits.open(filename)

        new_cube.spec_hdu = get_hdu(
            new_cube.hdul,
            hdu_index=spec_hdu_index,
            valid_names=KNOWN_SPEC_EXT_NAMES,
            msg_err_notfound=(
                "ERROR: Cannot determine which HDU contains spectral "
                "data, try to specify it manually!"
            ),
            msg_index_error="ERROR: Cannot open HDU {} to read specra!"
        )

        new_cube.var_hdu = get_hdu(
            new_cube.hdul,
            hdu_index=var_hdu_index,
            valid_names=KNOWN_VARIANCE_EXT_NAMES,
            msg_err_notfound=(
                "WARNING: Cannot determine which HDU contains the "
                "variance data, try to specify it manually!"
            ),
            msg_index_error="WARNING: Cannot open HDU {} to read the "
                            "variance!",
            exit_on_errors=False
        )

        new_cube.mask_hdu = get_hdu(
            new_cube.hdul,
            hdu_index=mask_hdu_index,
            valid_names=KNOWN_MASK_EXT_NAMES,
            msg_err_notfound=(
                "WARNING: Cannot determine which HDU contains the "
                "mask data, try to specify it manually!"
            ),
            msg_index_error="WARNING: Cannot open HDU {} to read the "
                            "mask!",
            exit_on_errors=False
        )

        new_cube.wd_hdu = get_hdu(
            new_cube.hdul,
            hdu_index=wd_hdu_index,
            valid_names=KNOWN_MASK_EXT_NAMES,
            msg_err_notfound=(
                "WARNING: Cannot determine which HDU contains the "
                "R data, try to specify it manually!"
            ),
            msg_index_error="WARNING: Cannot open HDU {} to read the "
                            "R matrix data!",
            exit_on_errors=False
        )
        return new_cube

    def write(self, filename: str, **kwargs):
        """
        Write the datacube if a FITS file.

        Parameters
        ----------
        filename : str
            The path of the output file.
        kwargs : dict
            Optional arguments passed to astropy.io.fits.HDUList.writeto.
            For example, it can be {'overwrite': True}.

        Returns
        -------
        bool
            True if the write operation was succesfull.

        """
        if self.hdul is None:
            return False
        self.hdul.writeto(filename, **kwargs)


def __cutout_argshandler(options=None):
    """
    Parse the arguments given by the user.

    Inputs
    ------
    options: list or None
        If none, args are parsed from the command line, otherwise the options
        list is used as input for argument parser.

    Returns
    -------
    args: Namespace
        A Namespace containing the parsed arguments. For more information see
        the python documentation of the argparse module.
    """
    parser = argparse.ArgumentParser(
        description='Generate cutouts of spectral cubes (or fits images, both '
        'grayscale or RGB).'
    )

    parser.add_argument(
        'input_fits', metavar='INPUT_FIST', type=str, nargs=1,
        help='The spectral cube (or image) from which to extract a cutout.'
    )
    parser.add_argument(
        '--regionfile', '-r', metavar='REGION_FILE', type=str, default=None,
        help='The region-file used to identify the locations and sizes of the '
        'cutouts. If multiple regions are present in the region-file, a cutout'
        ' is generated for each region. If the input file is a spectral '
        'datacube, the text field of the region can be used to specify an '
        'optional wavelength range for the cutout of that region. If a region '
        'do not provide a wavelength range information and the --wave-range '
        'option is specified, then the wavelength range specified by the '
        'latter parameter is used, otherwise the cutout will contain the full '
        'wavelength range as the original datacube. wavelength ranges are '
        'ignored for grayscale or RGB images.'
        'If this option is not specified, then the coordinate and the size of '
        'a cutout region must be specified with the options --center, --sizes '
        'and --wave-range.'
    )
    parser.add_argument(
        '--center', '-c', metavar='RA,DEC', type=str, default=None,
        help='Specify the RA and DEC of the center of a single cutout. '
        'Both RA and DEC can be specified with units in a format compatible '
        'with astropy.units (eg. -c 10arcsec,5arcsec). If no no unit is '
        'specified, then the quantity is assumed to be in pixels.'
    )
    parser.add_argument(
        '--size', '-s', metavar='WIDTH[,HEIGHT]', type=str, default=None,
        help='Specify the HEIGHT and WIDTH of a single cutout. If HEIGHT is '
        'not specified then it is assumed to be equal to WIDTH. '
        'Both HEIGHT and WIDTH can be specified with units in a format '
        'compatible with astropy.units (eg. -s 10arcsec,5arcsec).  If no no '
        'unit is specified, then the quantity is assumed to be in pixels.'
    )
    parser.add_argument(
        '--angle', '-t', metavar='ANGLE', type=str, default=0,
        help='Specify the rotation angle arounf the center of a single cutout.'
        '%(metavar)s can be specified with units in a format '
        'compatible with astropy.units (eg. -s 10arcsec,5arcsec).  If no no '
        'unit is specified, then the quantity is assumed to be in degrees.'
    )
    parser.add_argument(
        '--wave-range', '-w', metavar='MIN_W,MAX_W', type=str, default=None,
        help='Specify the wavelength range that the extracted cutout will '
        'contains. This option is ignored if the input file is a grayscale or '
        'an RGB image. Both HEIGHT and WIDTH can be specified with units in a '
        'format compatible with astropy.units '
        '(eg. -w 4500angstrom,6500angstrom). If no no unit is specified, then '
        'the quantity is assumed to be in angstrom.'
    )
    parser.add_argument(
        '--data-hdu', metavar='DATA_HDU[,HDU1,HDU2]', type=str,
        default=None, help='Specify which extensions contain data. For an '
        'rgb image more than one HDU can be specified, for example '
        '--data-hdu 1,2,3. If this option is not specified then the program '
        'will try to identify the data type and structure automatically.'
    )
    parser.add_argument(
        '--verbose', action='store_true', default=False,
        help="Print verbose outout."
    )
    args = None
    if options is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(options)

    inp_fits_file = args.input_fits[0]
    if (inp_fits_file is not None) and (not os.path.isfile(inp_fits_file)):
        print(f"The file {inp_fits_file} does not exist!")
        sys.exit(1)

    if args.regionfile is None:
        if (args.center is None) or (args.size is None):
            print(
                "If --regionfile is not specified then both --center and "
                "--size must be provided."
            )
            sys.exit(1)
    elif not os.path.isfile(args.regionfile):
        print("The file input regionfile does not exist!")
        sys.exit(1)

    return args


def __smoothing_argshandler(options=None):
    """
    Parse the arguments given by the user.

    Inputs
    ------
    options: list or None
        If none, args are parsed from the command line, otherwise the options
        list is used as input for argument parser.

    Returns
    -------
    args: Namespace
        A Namespace containing the parsed arguments. For more information see
        the python documentation of the argparse module.
    """
    parser = argparse.ArgumentParser(
        description='Apply a gaussian smoothing kernel to a spectral cubes '
                    'spatially and/or along the spectral axis.'
    )

    parser.add_argument(
        'input_fits_files', metavar='INPUT_FIST', type=str, nargs='+',
        help='The spectral cube (or image) from which to extract a cutout.'
    )
    parser.add_argument(
        '--spatial-sigma', metavar='SIGMA', type=str, default=1.0,
        help='Set the sigma for the spatial gaussian kernel. If %(metavar)s '
        'is 0 then no spatial smoothing is applied. If not specified the '
        'default value %(metavar)s=%(default)f is used.'
    )
    parser.add_argument(
        '--spatial-supersample', metavar='ZOOM_FACTOR', type=int, default=0,
        help='Set the spatial supersampling factor. If %(metavar)s <= 1 then '
        'no supersampling is applied. %(metavar)s=2 means that the output cube'
        ' will have a doubled width and height, and so on. The default value '
        'is %(metavar)s=%(default)d.'
    )
    parser.add_argument(
        '--wave-supersample', metavar='ZOOM_FACTOR', type=int, default=0,
        help='Set the wavelength supersampling factor. If %(metavar)s <= 1 '
        'then no supersampling is applied. %(metavar)s=2 means that the output'
        'cube will have a doubled spectral resolution, and so on. '
        'The default value is %(metavar)s=%(default)d.'
    )
    parser.add_argument(
        '--wave-sigma', metavar='SIGMA', type=str, default=0,
        help='Set the sigma for the spectral gaussian kernel. If %(metavar)s '
        'is 0 then no spectral smoothing is applied. If not specified the '
        'default value %(metavar)s=%(default)f is used.'
    )
    parser.add_argument(
        '--info-hdu', metavar='INFO_HDU', type=int, default=0,
        help='The HDU containing cube metadata. If this argument '
        'Set this to -1 to automatically detect the HDU containing the info. '
        'NOTE that this value is zero indexed (i.e. firts HDU has index 0).'
    )

    parser.add_argument(
        '--spec-hdu', metavar='SPEC_HDU', type=int, default=-1,
        help='The HDU containing the spectral data to use. If this argument '
        'Set this to -1 to automatically detect the HDU containing spectra. '
        'NOTE that this value is zero indexed (i.e. second HDU has index 1).'
    )

    parser.add_argument(
        '--var-hdu', metavar='VAR_HDU', type=int, default=-1,
        help='The HDU containing the variance of the spectral data. '
        'Set this to -1 if no variance data is present in the cube. '
        'The default value is %(metavar)s=%(default)s.'
        'NOTE that this value is zero indexed (i.e. third HDU has index 2).'
    )

    parser.add_argument(
        '--mask-hdu', metavar='MASK_HDU', type=int, default=-1,
        help='The HDU containing the valid pixel mask of the spectral data. '
        'Set this to -1 if no mask is present in the cube. '
        'The default value is %(metavar)s=%(default)s.'
        'NOTE that this value is zero indexed (i.e. fourth HDU has index 3).'
    )

    parser.add_argument(
        '--verbose', action='store_true', default=False,
        help="Print verbose outout."
    )

    parser.add_argument(
        '--debug', action='store_true', default=False,
        help="Start an IPython console for debugging."
    )

    args = None
    if options is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(options)

    for inp_fits_file in args.input_fits_files:
        if (inp_fits_file is not None) and (not os.path.isfile(inp_fits_file)):
            print(f"The file {inp_fits_file} does not exist!")
            sys.exit(1)

    if (args.spatial_sigma == 0) and (args.wave_sigma == 0):
        print(
            "Spatial smoothing and spectral smooting cannot be both disabled, "
            "please set at least one the options --wave-sigma or "
            "--spatial-sigma."
        )
        sys.exit(1)

    return args


def get_gray_cutout(
    data: np.ndarray,
    center: Union[SkyCoord, tuple, list],
    size: Union[tuple, list],
    angle: Optional[Union[float, units.Quantity]] = 0,
    data_wcs: Optional[WCS] = None
) -> Dict[str, Union[WCS, np.ndarray]]:
    """
    Get the cutout for a grayscale image.

    This is a basic wrapper around astropy.nddata.utils.Cutout2D

    Parameters
    ----------
    data : np.ndarray
        The actual image data. Should have only two dimensions (a grayscale
        image has only X and Y corrdinates).
    center : astropy.coordinate.SkyCoord or tuple.
        The center of the cutout. If a SkyCoord is provided then a WCS for the
        image data musto also be specified with the parameter data_wcs.
        If a tuple is provided, then the first two numbers of the tuple are
        interpreted as the X and Y coordinate of the cutout center: in this
        case, if no WCS is specified, the values are assumed to be in pixels,
        else if a WCS is provided then the values are assumed to be in degrees.
    size : tuple or list
        The first two values in the tuple are interpreted as the width and
        height of the cutout. if no WCS is specified, the values are assumed to
        be in pixels, else if a WCS is provided then the values are assumed to
        be in degrees. Astropy.units.Quantity values are also supported.
    angle : float or astropy.units.Quantity, optional
        The rotation angle of the cutout. If it is a float, then it is
        interpreted in degrees. The default is 0.
    data_wcs : astropy.wcs.WCS or None, optional
        A WCS associated with the image data. The default is None.

    Returns
    -------
    cutout_dict: dict
        A dictionary containing the following key: value pairs:
            'data': np.ndarray
                The cutout data.
            'wcs': astropy.wcs.WCS or None
                The wcs for the cutout data.
    """
    angle = units.Quantity(angle, units.deg)
    if data_wcs is not None:
        data_wcs = data_wcs.celestial
        sx, sy, rot, shr_y = get_pc_transform_params(data_wcs)
        angle = angle - rot

    rotated_data = rotate_data(
        data=data,
        angle=angle,
        data_wcs=data_wcs
    )

    cutout = Cutout2D(
        rotated_data['data'],
        center, size,
        mode='partial',
        fill_value=np.nan,
        wcs=rotated_data['wcs'],
        copy=True
    )

    cutout_dict = {
        'data': cutout.data,
        'wcs': cutout.wcs
    }

    return cutout_dict


def get_rgb_cutout(
    data: Union[tuple, list, np.ndarray],
    center: Union[SkyCoord, tuple],
    size: Union[tuple, list],
    angle: Optional[Union[float, units.Quantity]] = 0,
    data_wcs: Optional[Union[WCS, list, tuple]] = None,
    resample_to_wcs: bool = False
) -> Dict[
    str,
    Union[
        Tuple[WCS, WCS, WCS],
        Tuple[np.ndarray, np.ndarray, np.ndarray]
    ]
]:
    """
    Get a cutout from a bigger RGB.

    Parameters
    ----------
    data : np.ndarray or tuple or list
        The actual image data.
    center : astropy.coordinate.SkyCoord or tuple.
        The center of the cutout. If a SkyCoord is provided then a WCS for the
        image data musto also be specified with the parameter data_wcs.
        If a tuple is provided, then the first two numbers of the tuple are
        interpreted as the X and Y coordinate of the cutout center: in this
        case, if no WCS is specified, the values are assumed to be in pixels,
        else if a WCS is provided then the values are assumed to be in degrees.
    size : tuple
        The first two values in the tuple are interpreted as the width and
        height of the cutout. if no WCS is specified, the values are assumed to
        be in pixels, else if a WCS is provided then the values are assumed to
        be in degrees. Astropy.units.Quantity values are also supported.
    angle : float or astropy.units.Quantity, optional
        The rotation angle of the cutout. If it is a float, then it is
        interpreted in degrees. The default is 0.
    data_wcs : astropy.wcs.WCS or None, optional
        A WCS associated with the image data. The default is None.
    resample_to_wcs : bool, optional
        If true reample the red, green and blue data to share the same WCS.
        In order to use this option, the WCSs for the input data must be
        provided, otherwise this option will be ignored and a warning message
        is outputed. The default is False.

    Returns
    -------
    cutout_dict: dict
        A dictionary containing the following key: value pairs:
            'data': np.ndarray
                The cutout data.
            'wcs': astropy.wcs.WCS or None
                The wcs for the cutout data.
    """
    # Do some sanity checks on the input parameters
    if isinstance(data, np.ndarray):
        if len(data.shape) != 3 or data.shape[2] != 3:
            raise ValueError(
                "Only RGB images are supported: expected shape (N, M, 3) but"
                f"input data has shape {data.shape}."
            )
        if data_wcs is not None:
            data_wcs_r = data_wcs.celestial
            data_wcs_g = data_wcs.celestial
            data_wcs_b = data_wcs.celestial

        data_r = data[..., 0]
        data_g = data[..., 1]
        data_b = data[..., 2]
    elif isinstance(data, Union[tuple, list]):
        if len(data) != 3:
            raise ValueError(
                "'data' parameter only accepts list or tuple containing "
                "exactly 3 elements."
            )
        elif not all([isinstance(x, np.ndarray) for x in data]):
            raise ValueError(
                "All elements of the input tupel or list must be 2D arrays."
            )
        if data_wcs is None:
            if resample_to_wcs:
                warnings.warn(
                    "reample_to_wcs is set to True but no WCS info is provided"
                )
                resample_to_wcs = False
            resample_to_wcs = False
            data_wcs_r = None
            data_wcs_g = None
            data_wcs_b = None
        else:
            if not isinstance(data_wcs, Union[tuple, list]):
                raise ValueError(
                    "When 'data' is a list or a tuple, also data_wcs must be a"
                    "a list or a tuple of WCSs."
                )
            elif not all([isinstance(x, WCS) for x in data_wcs]):
                raise ValueError(
                    "All elements of wcs_data tuple or list must be WCS."
                )
            data_wcs_r = data_wcs[0].celestial
            data_wcs_g = data_wcs[1].celestial
            data_wcs_b = data_wcs[2].celestial
        data_r = data[0]
        data_g = data[1]
        data_b = data[2]
    else:
        raise ValueError(
            "Parameter 'data' only supports ndarray or list/tuple of ndarrays."
        )

    cutout_data_r = get_gray_cutout(data_r, center, size, angle, data_wcs_r)
    cutout_data_g = get_gray_cutout(data_g, center, size, angle, data_wcs_g)
    cutout_data_b = get_gray_cutout(data_b, center, size, angle, data_wcs_b)

    if not resample_to_wcs:
        cutout_dict = {
            'data': (
                cutout_data_r['data'],
                cutout_data_g['data'],
                cutout_data_b['data']
            ),
            'wcs': (
                cutout_data_r['wcs'],
                cutout_data_g['wcs'],
                cutout_data_b['wcs'],
            )
        }

    return cutout_dict


def get_cube_cutout(
    data: np.ndarray,
    center: Union[SkyCoord, tuple, list],
    size: Union[tuple, list],
    angle: Optional[Union[float, units.Quantity]] = 0,
    wave_range: Optional[Union[tuple, list]] = None,
    data_wcs: Optional[WCS] = None,
    report_callback: Optional[Callable] = None
) -> Dict[str, Union[WCS, np.ndarray]]:
    """
    Get a cutout of a spectral datacube.

    Parameters
    ----------
    data : np.ndarray
        The datacube data.
    center : astropy.coordinate.SkyCoord or tuple or list.
        The center of the cutout. If a SkyCoord is provided then a WCS for the
        image data musto also be specified with the parameter data_wcs.
        If a tuple is provided, then the first two numbers of the tuple are
        interpreted as the X and Y coordinate of the cutout center: in this
        case, if no WCS is specified, the values are assumed to be in pixels,
        else if a WCS is provided then the values are assumed to be in degrees.
    size : tuple or list
        The first two values in the tuple are interpreted as the width and
        height of the cutout. Both adimensional values and angular quantities
        are accepted. Adimensional values are interpreted as pixels.
        Angular values are converted to pixel values ignoring any non-linear
        distorsion.
    angle : float or astropy.units.Quantity, optional
        The rotation angle of the cutout. If it is a float, then it is
        interpreted in degrees. The default is 0.
    wave_range : tuple or list, optional
        If not None, he first two values in the tuple are interpreted as the
        minimum and maximum value of the wavelength range for the cutout.
        If it is None, the whole wavelength range is used. The default is None.
    data_wcs : astropy.wcs.WCS or None, optional
        A WCS associated with the image data. The default is None..
    report_callback : Callable or None, optional
        A callable that will be executed every time the cutout of a single
        slice of the cube is computed. Must accept in input two arguments:

          * the number of slice processed so far
          * the total number of slices.

    Returns
    -------
    cutout_dict
        A dictionary containing the following key: value pairs:
            'data': np.ndarray
                The cutout data.
            'wcs': astropy.wcs.WCS or None
                The wcs for the cutout data.
    """
    # Do some sanity checks on the input data
    if len(data.shape) != 3:
        raise ValueError("Unsupported datacube shape {data.shape}.")

    if not isinstance(size, Union[list, tuple]):
        raise ValueError(
            "'size' must be a list or a tuple of scalar values or angular "
            "quantities"
        )
    elif not all(
        [isinstance(x, Union[int, float, units.Quantity]) for x in size]
    ):
        raise ValueError(
            "'size' must be a list or a tuple of scalar values or angular "
            "quantities"
        )

    d_a, d_b = size[:2]

    if not isinstance(center, Union[SkyCoord, tuple, list]):
        raise ValueError("'center' must be SkyCoord or tuple or list.")

    angle = units.Quantity(angle, units.deg)

    if data_wcs is not None:
        if not isinstance(data_wcs, WCS):
            raise ValueError(
                "'data_wcs' must be eihter None or a valid WCS object"
            )

        if not data_wcs.has_spectral:
            raise ValueError(
                "The provided WCS does not seem to have a spectral axis"
            )

        celestial_wcs = data_wcs.celestial
        specex_wcs = data_wcs.spectral
    else:
        celestial_wcs = None
        specex_wcs = None

    cutout_data = []
    for k in range(data.shape[0]):
        cutout = get_gray_cutout(
            data=data[k],
            center=center,
            size=size,
            angle=angle,
            data_wcs=celestial_wcs
        )

        cutout_data.append(cutout['data'])
        if report_callback is not None:
            report_callback(k, data.shape[0]-1)

    cutout_data = np.array(cutout_data)

    if celestial_wcs is not None:
        wcs_header = cutout['wcs'].to_header()
        wcs_header['CRPIX3'] = specex_wcs.wcs.crpix[0]
        wcs_header['PC3_3'] = specex_wcs.wcs.get_pc()[0, 0]
        wcs_header['PC1_3'] = 0
        wcs_header['PC2_3'] = 0
        wcs_header['PC3_2'] = 0
        wcs_header['PC3_1'] = 0
        wcs_header['CDELT3'] = specex_wcs.wcs.cdelt[0]
        wcs_header['CUNIT3'] = str(specex_wcs.wcs.cunit[0])
        wcs_header['CTYPE3'] = specex_wcs.wcs.ctype[0]
        wcs_header['CRVAL3'] = specex_wcs.wcs.crval[0]

    else:
        wcs_header = None

    return {
        'data': cutout_data,
        'wcs': WCS(wcs_header)
    }


def _get_fits_data_structure(fits_file: str) -> Dict[str, Any]:
    data_structure = {
        'type': None,
        'data-ext': [],
        'variance-ext': [],
        'mask-ext': []
    }
    with fits.open(fits_file) as f:
        # If there is only one extension, than it should contain the image data
        if len(f) == 1:
            data_ext = f[0]
            data_structure['data-ext'] = [0, ]
        else:
            # Otherwise, try to identify the extension form its name
            for k, ext in enumerate(f):
                if ext.name.lower() in KNOWN_SPEC_EXT_NAMES:
                    data_ext = ext
                    data_structure['data-ext'] = [k, ]
                    break

                if ext.name.lower() in KNOWN_RGB_EXT_NAMES:
                    data_ext = ext
                    data_structure['data-ext'].append(k)

            # If cannot determine which extensions cointain data,
            # then just use the second extension
            if not data_structure['data-ext']:
                data_ext = f[1]
                data_structure['data-ext'] = [1, ]

        data_shape = data_ext.data.shape
        if len(data_shape) == 2:
            # A 2D image, we should check other extensions to
            # determine if its an RGB multi-extension file
            for k, ext in enumerate(f):
                if k in data_structure['data-ext']:
                    continue

                lower_ext_name = ext.name.strip().lower()

                if ext.data is not None and ext.data.shape == data_shape:
                    if lower_ext_name:
                        if (
                                lower_ext_name in KNOWN_SPEC_EXT_NAMES or
                                lower_ext_name in KNOWN_RGB_EXT_NAMES
                        ):
                            data_structure['data-ext'].append(k)
                        elif lower_ext_name in KNOWN_VARIANCE_EXT_NAMES:
                            data_structure['variance-ext'].append(k)
                        elif lower_ext_name in KNOWN_MASK_EXT_NAMES:
                            data_structure['mask-ext'].append(k)
                        else:
                            continue
                    else:
                        data_structure['data-ext'].append(k)

            if len(data_structure['data-ext']) == 1:
                data_structure['type'] = 'image-gray'
            elif len(data_structure['data-ext']) == 3:
                data_structure['type'] = 'image-rgb'
            else:
                data_structure['type'] = 'unkown'

        elif len(data_shape) == 3:
            # Could be a datacube or an RGB cube or a weird grayscale image,
            # depending on the size of third axis. Only grayscale image will be
            # treated separately, while an RGB cube will be treated as a normal
            # datacube
            if data_shape[2] == 1:
                # A very weird grayscale image.
                data_structure['type'] = 'cube-gray'
            else:
                data_structure['type'] = 'cube'
                for k, ext in enumerate(f):
                    ext_name = ext.name.strip().lower()

                    if k in data_structure['data-ext']:
                        continue
                    elif ext_name in KNOWN_VARIANCE_EXT_NAMES:
                        data_structure['variance-ext'].append(k)
                    elif ext_name in KNOWN_MASK_EXT_NAMES:
                        data_structure['mask-ext'].append(k)

        else:
            # We dont know how to handle weird multidimensional data.
            print(
                "WARNING: cannot handle multidimensional data with shape "
                f"{data_shape}"
            )
            data_structure['type'] = 'unkown'

    if not data_structure['data-ext']:
        data_structure['data-ext'] = None
    if not data_structure['variance-ext']:
        data_structure['variance-ext'] = None
    if not data_structure['mask-ext']:
        data_structure['mask-ext'] = None

    return data_structure


def get_hdu(hdl, valid_names, hdu_index=None, msg_err_notfound=None,
            msg_index_error=None, exit_on_errors=True):
    """
    Find a valid HDU in a HDUList.

    Parameters
    ----------
    hdl : list of astropy.io.fits HDUs
        A list of HDUs.
    valid_names : list or tuple of str
        A list of possible names for the valid HDU.
    hdu_index : int or str, optional
        Manually specify which HDU to use. The default is None.
    msg_err_notfound : str or None, optional
        Error message to be displayed if no valid HDU is found.
        The default is None.
    msg_index_error : str or None, optional
        Error message to be displayed if the specified index is outside the
        HDU list boundaries.
        The default is None.
    exit_on_errors : bool, optional
        If it is set to True, then exit the main program with an error if a
        valid HDU is not found, otherwise just return None.
        The default value is True.

    Returns
    -------
    valid_hdu : astropy.io.fits HDU or None
        The requested HDU.

    """
    valid_hdu = None
    if hdu_index is None:
        # Try to detect HDU containing spectral data
        for hdu in hdl:
            if hdu.name.lower() in valid_names:
                valid_hdu = hdu
                break
        else:
            if msg_err_notfound:
                print(msg_err_notfound, file=sys.stderr)
            if exit_on_errors:
                sys.exit(1)
    else:
        try:
            valid_hdu = hdl[hdu_index]
        except IndexError:
            if msg_index_error:
                print(msg_index_error.format(hdu_index), file=sys.stderr)
            if exit_on_errors:
                sys.exit(1)
    return valid_hdu


def cube_tiled_func(data, func, tile_size, *args, **kwargs):
    data_shape = data.shape[-2:]
    if isinstance(data, np.ma.MaskedArray):
        result = np.ma.zeros(data_shape)
    else:
        result = np.zeros(data_shape)
    for j in np.arange(data_shape[0], step=tile_size):
        for k in np.arange(data_shape[1], step=tile_size):
            tile = data[:, j:j+tile_size, k:k+tile_size]
            # Skip empty tiles:
            if not np.isfinite(tile).any():
                result[j:j+tile_size, k:k+tile_size] = np.nan
                try:
                    result[j:j+tile_size, k:k+tile_size].mask = True
                except AttributeError:
                    pass
                continue

            processed_tile = func(tile, *args, **kwargs).copy()
            result[j:j+tile_size, k:k+tile_size] = processed_tile
            try:
                result[j:j+tile_size, k:k+tile_size].mask = processed_tile.mask
            except AttributeError:
                pass

    return result


def correlate_spaxel(cube_data: np.ndarray,
                     spaxel_data: np.ndarray,
                     similarity_function: Optional[str] = 'rms'):

    if spaxel_data.shape[0] != cube_data.shape[0]:
        raise ValueError(
            "spaxel_data and cube_data must have the same first dimension."
        )

    if len(spaxel_data.shape) == 1:
        spaxel_data = spaxel_data[:, None, None]

    x = cube_data - np.nanmedian(cube_data, axis=0)
    x = x / np.nanmax(x, axis=0)
    y = spaxel_data - np.nanmedian(spaxel_data)
    y = y / np.nanmax(y)

    if similarity_function == 'rms':
        res = np.sqrt(np.nanmean((x - y)**2, axis=0))
        return 1/(1 + res)
    elif similarity_function == 'correlation':
        res = np.nansum(x * y, axis=0)
        return res / (np.nansum(x**2, axis=0) * np.nansum(y**2))


def get_continuum_subtracted_slice(
        data: np.ndarray,
        line_wave: Union[int, units.Quantity],
        line_window: Union[int, units.Quantity] = 10 * units.angstrom,
        continuum_window: Union[int, units.Quantity] = 10 * units.angstrom,
        variance: Optional[np.ndarray] = None,
        data_mask: Optional[np.ndarray] = None,
        cube_wcs: WCS = None
) -> np.ndarray:
    """
    Get a continuum subtracted image from a spectral datacube.

    Parameters
    ----------
    data : np.ndarray
        DESCRIPTION.
    line_wave : Union[int, units.Quantity]
        DESCRIPTION.
    line_window : Union[int, units.Quantity], optional
        DESCRIPTION. The default is 10 * units.angstrom.
    continuum_window : Union[int, units.Quantity], optional
        DESCRIPTION. The default is 10 * units.angstrom.
    variance : Optional[np.ndarray], optional
        DESCRIPTION. The default is None.
    data_mask : Optional[np.ndarray], optional
        DESCRIPTION. The default is None.
    cube_wcs : WCS, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    None.

    """
    param_is_dimensional = [
        isinstance(x, units.Quantity)
        for x in (line_wave, line_window, continuum_window)
    ]

    if any(param_is_dimensional):
        if not all(param_is_dimensional):
            raise ValueError(
                "central_wave, line_window and continuum_window must be all "
                "integer indices or all dimensional quantities."
            )
        elif cube_wcs is None:
            raise ValueError(
                "A valid WCS object must be provided when central_wave, "
                "line_window and continuum_window are dimensional quantities."
            )
        line_wave_pix = cube_wcs.spectral.world_to_pixel(line_window)

        line_window_pix_low = cube_wcs.spectral.world_to_pixel(
            line_wave - line_window
        )

        line_window_pix_high = cube_wcs.spectral.world_to_pixel(
            line_wave + line_window
        )

    else:
        line_wave_pix = line_window
        line_window_pix_low = int(line_wave_pix - line_window)
        line_window_pix_high = int(line_wave_pix + line_window)

        """
        continuum_window_pix_low = int(
            line_window_pix_low - continuum_window/2
        )
        continuum_window_pix_high = int(
            line_window_pix_high + continuum_window/2
        )

        cx_low = (continuum_window_pix_low + line_window_pix_low) / 2
        cy_low = data[
            continuum_window_pix_low:line_window_pix_low, ...
        ].mean()

        cx_high = (line_window_pix_low + continuum_window_pix_high) / 2
        cy_high = data[
            continuum_window_pix_low:line_window_pix_low, ...
        ].mean()
        """

    line_slice = data[line_window_pix_low:line_window_pix_high, ...]


def self_correlate(data: np.ndarray,
                   data_mask: Optional[np.ndarray] = None,
                   similarity_sigma_threshold: Optional[float] = 5,
                   tile_size: Optional[int] = 32,
                   block_size: Optional[int] = 2,
                   similarity_function: Optional[str] = 'rms',
                   report_callback: Optional[Callable] = None) -> np.ndarray:
    if data_mask is not None and data.shape != data_mask.shape:
        raise ValueError("data and data_mask must have the same shape!")

    hei, wid = data.shape[1:]

    sim_table = np.zeros((hei, wid))
    # For each spaxel in the cube
    block_id = 0
    for h in np.arange(hei, step=block_size):
        for k in np.arange(wid, step=block_size):
            if report_callback is not None:
                report_callback(block_id, int(wid*hei / (block_size**2))-1)

            block_id += 1
            if (
                (sim_table[h:h+block_size, k:k+block_size] != 0).any() or
                (
                    data_mask is not None and
                    (data_mask[:, h:h+block_size, k:k+block_size] != 0).any()
                )
            ):
                continue

            spaxel_data = np.nansum(
                data[:, h:h+block_size, k:k+block_size],
                axis=(1, 2)
            )

            if not np.isfinite(spaxel_data).any():
                continue

            similarity_map = cube_tiled_func(
                data,
                correlate_spaxel,
                tile_size=tile_size,
                spaxel_data=spaxel_data,
                similarity_function=similarity_function
            )

            thresh = np.nanmean(similarity_map)
            thresh += similarity_sigma_threshold * np.nanstd(similarity_map)
            similarity_mask = similarity_map >= thresh
            sim_table[similarity_mask] = block_id
    return sim_table


def smooth_cube(data: np.ndarray, data_mask: Optional[np.ndarray] = None,
                spatial_sigma: Optional[Union[float, units.Quantity]] = 1.0,
                wave_sigma: Optional[Union[float, units.Quantity]] = 0.0,
                cube_wcs : Optional[WCS] = None,
                report_callback: Optional[Callable] = None) -> np.ndarray:
    """
    Smooth a datacube spatially and/or along the spectral axis.

    Parameters
    ----------
    data : numpy.ndarray
        The spectral datacube.
    data_mask : numpy.ndarray, optional
        The mask for the spectral datacube. The default is None.
    spatial_sigma : float, optional
        The sigma for the spatial smoothing gaussian kernel.
        The default is 1.0.
    wave_sigma : float, optional
        The sigma fot the spectral smoothing gaussian kernel.
        The default is 0.0.
    report_callback : Callable or None, optional
        A callable that will be execute every time the cutout of a single
        slice of the cube is computed. Must accept in input two arguments:

          * the number of slice processed so far
          * the total number of slices.

    Raises
    ------
    ValueError
        If the shape of data does not match the shape of data_mask.

    Returns
    -------
    smoothed_arr : numpy.ndarray
        The smoothed version of the input data.

    """
    if data_mask is not None and data.shape != data_mask.shape:
        raise ValueError("data and data_mask must have the same shape!")

    smoothed_arr = data.copy()

    if not isinstance(wave_sigma, Number):
        if cube_wcs is None:
            raise ValueError(
                "cube_wcs is needed when using dimensional quantities "
                "for wave_sigma"
            )
        elif not cube_wcs.has_spectral:
            raise ValueError(
                "Datacube WCS has no spectral axis!"
            )
        else:
            # Determine the equivalent sigma in pixel space
            p0 = cube_wcs.spectral.pixel_to_world(0).to(units.angstrom)
            wave_sigma = cube_wcs.spectral.world_to_pixel(p0 + wave_sigma)
    do_wave_smoohting = wave_sigma > 0

    if (
        isinstance(spatial_sigma, Number) or
        isinstance(spatial_sigma[0], Number)
    ):
        try:
            dp_spatial_smoothing = spatial_sigma > 0
        except TypeError:
            dp_spatial_smoothing = spatial_sigma[0] > 0
            dp_spatial_smoothing |= spatial_sigma[1] > 0
    else:
        if cube_wcs is None:
            raise ValueError(
                "cube_wcs is needed when using dimensional quantities "
                "for spatial_sigma"
            )
        elif not cube_wcs.has_celestial:
            raise ValueError(
                "Datacube WCS has no celestial axes!"
            )
        else:
            # Determine the equivalent sigma in pixel space
            p0 = cube_wcs.celestial.pixel_to_world(0, 0)
            try:
                d_ra = spatial_sigma[0]
                d_dec = spatial_sigma[1]
            except (TypeError, IndexError):
                d_ra = spatial_sigma
                d_dec = spatial_sigma
            p1 = p0.spherical_offsets_by(d_ra, d_dec)
            spatial_sigma = cube_wcs.celestial.world_to_pixel(p1)
            dp_spatial_smoothing = (spatial_sigma[0] > 0)
            dp_spatial_smoothing |= (spatial_sigma[1] > 0)

    if do_wave_smoohting:
        for h in range(smoothed_arr.shape[1]):
            if report_callback is not None:
                report_callback(h, smoothed_arr.shape[1] - 1)
            for k in range(smoothed_arr.shape[2]):
                smoothed_spaxel = gaussian_filter1d(
                    smoothed_arr[:, h, k],
                    sigma=wave_sigma,
                    mode='constant'
                )
                smoothed_arr[:, h, k] = smoothed_spaxel

    if report_callback is not None:
        print("")

    if dp_spatial_smoothing:
        for k, data_slice in enumerate(smoothed_arr):
            if report_callback is not None:
                report_callback(k, smoothed_arr.shape[0] - 1)
            smoothed_slice = gaussian_filter(
                data_slice,
                sigma=spatial_sigma,
                mode='constant'
            )
            smoothed_arr[k] = smoothed_slice

    if report_callback is not None:
        print("")

    if data_mask is not None:
        smoothed_mask = data_mask.copy().astype(bool)
        for k, mask_slice in enumerate(smoothed_mask):
            if report_callback is not None:
                report_callback(k, smoothed_mask.shape[0] - 1)
                smoothed_mask[k] &= ~np.isfinite(smoothed_arr[k])
        if report_callback is not None:
            print("")
        smoothed_mask = smoothed_mask.astype('int8')
    else:
        smoothed_mask = None

    return smoothed_arr, smoothed_mask


def smoothing_main(options=None):
    """
    Run the main cutout program.

    Parameters
    ----------
    options : list or None, optional
        A list of cli input prameters. The default is None.

    Returns
    -------
    None.

    """
    args = __smoothing_argshandler(options)

    if args.verbose:
        report_callback = simple_pbar_callback
    else:
        report_callback = None

    try:
        wave_sigma = float(args.wave_sigma)
    except ValueError:
        wave_sigma = units.Quantity(args.wave_sigma)

    spatial_sigma_list = args.spatial_sigma.split(',')
    try:
        if len(spatial_sigma_list) > 1:
            spatial_sigma = (
                float(spatial_sigma_list[0]),
                float(spatial_sigma_list[1])
            )
        else:
            spatial_sigma = float(spatial_sigma_list[0])
    except ValueError:
        if len(spatial_sigma_list) > 1:
            spatial_sigma = (
                units.Quantity(spatial_sigma_list[0]),
                units.Quantity(spatial_sigma_list[1])
            )
        else:
            spatial_sigma = units.Quantity(spatial_sigma_list[0])

    for target_data_file in args.input_fits_files:
        fits_base_name = os.path.basename(target_data_file)
        fits_base_name = os.path.splitext(fits_base_name)[0]
        out_fname = f"{fits_base_name}_smoothed.fits"

        if args.verbose:
            print(f"\n[{fits_base_name}]")

        shutil.copy(target_data_file, out_fname)

        with SpectralCube.open(target_data_file, mode='readonly') as my_cube:

            data_mask = None
            if my_cube.mask_hdu is not None:
                data_mask = my_cube.mask_hdu.data

            if args.verbose:
                print(">>> applying smoothing...")
                print(f"  - spatial_sigma: {spatial_sigma}")
                print(f"  - wave_sigma: {wave_sigma}")

            smoothed_spec, smoothed_mask = smooth_cube(
                data=my_cube.spec_hdu.data,
                data_mask=data_mask,
                spatial_sigma=spatial_sigma,
                wave_sigma=wave_sigma,
                cube_wcs=my_cube.getSpecWCS(),
                report_callback=report_callback
            )
            my_cube.spec_hdu.data = smoothed_spec

            if my_cube.var_hdu is not None:
                smoothed_var, _ = smooth_cube(
                    data=my_cube.var_hdu.data,
                    spatial_sigma=spatial_sigma,
                    wave_sigma=wave_sigma,
                    cube_wcs=my_cube.getVarWCS(),
                    report_callback=report_callback
                )
                my_cube.var_hdu.data = smoothed_var

            if my_cube.mask_hdu is not None:
                my_cube.mask_hdu.data = smoothed_mask

            if args.debug:
                embedder = get_ipython_embedder()
                if embedder is not None:
                    embedder()

            if args.verbose:
                print(f"  - saving to {out_fname}...")

            my_cube.write(
                out_fname,
                overwrite=True
            )


def cutout_main(options=None):
    """
    Run the main cutout program.

    Parameters
    ----------
    options : list or None, optional
        A list of cli input prameters. The default is None.

    Returns
    -------
    None.

    """
    def updated_wcs_cutout_header(orig_header, cutout_header):
        new_header = orig_header.copy()

        sys.stdout.flush()
        sys.stderr.flush()

        # Delete any existing CD card
        cd_elem_re = re.compile(r'CD[1-9]_[1-9]')
        for k in list(new_header.keys()):
            if cd_elem_re.fullmatch(str(k).strip()):
                new_header.remove(k, ignore_missing=True, remove_all=True)

        # Delete any existing PC card
        pc_elem_re = re.compile(r'PC[1-9]_[1-9]')
        for k in list(new_header.keys()):
            if pc_elem_re.fullmatch(str(k).strip()):
                new_header.remove(k, ignore_missing=True, remove_all=True)

        # Copy new PC cards into the new header
        for k in list(cutout_header.keys()):
            if pc_elem_re.fullmatch(str(k).strip()):
                new_header[k] = cutout_header[k]

        new_header['PC1_1'] = cutout_header['PC1_1']
        new_header['PC2_2'] = cutout_header['PC2_2']
        new_header['CDELT1'] = cutout_header['CDELT1']
        new_header['CDELT2'] = cutout_header['CDELT2']
        new_header['CRVAL1'] = cutout_header['CRVAL1']
        new_header['CRVAL2'] = cutout_header['CRVAL2']
        new_header['CRPIX1'] = cutout_header['CRPIX1']
        new_header['CRPIX2'] = cutout_header['CRPIX2']
        new_header['CUNIT1'] = cutout_header['CUNIT1']
        new_header['CUNIT2'] = cutout_header['CUNIT2']

        try:
            crval3 = units.Quantity(
                cutout_header['CRVAL3'],
                cutout_header['CUNIT3']
            )
            cutout_unit3 = units.Quantity(1, cutout_header['CUNIT3'])
        except KeyError:
            pass
        else:
            c_factor = cutout_unit3.to(orig_header['CUNIT3']).value
            new_header['PC3_3'] = cutout_header['PC3_3'] * c_factor
            new_header['PC1_3'] = 0
            new_header['PC2_3'] = 0
            new_header['PC3_1'] = 0
            new_header['PC3_2'] = 0
            new_header['CDELT3'] = cutout_header['CDELT3']
            new_header['CRVAL3'] = crval3.to(orig_header['CUNIT3']).value
            new_header['CUNIT3'] = orig_header['CUNIT3']

        return new_header

    args = __cutout_argshandler(options)

    cutout_list = []
    if args.regionfile is not None:
        if HAS_REGION:
            myt, _ = parse_regionfile(
                args.regionfile,  key_ra='RA', key_dec='DEC'
            )
        else:
            logging.error("Astropy Regions is needed to handle regionfiles!")
            sys.exit(1)
        for row in myt:
            reg = row['region']

            if isinstance(reg, regions.CircleSkyRegion):
                center = reg.center
                sizes = [reg.radius*2, reg.radius*2]
                angle = 0
            elif (
                isinstance(reg, regions.EllipseSkyRegion) or
                isinstance(reg, regions.RectangleSkyRegion)
            ):
                center = reg.center
                sizes = [reg.height, reg.width]
                angle = reg.angle
            elif isinstance(reg, regions.EllipseAnnulusSkyRegion):
                center = reg.center
                sizes = [reg.outer_height, reg.outer_width]
                angle = reg.angle
            elif isinstance(reg, regions.PolygonSkyRegion):
                center = SkyCoord(
                    row['RA'],
                    row['DEC'],
                    unit='deg',
                    frame=reg.vertices[0].frame
                )
                ra_list = [x.ra.to('deg').value for x in reg.vertices]
                dec_list = [x.dec.to('deg').value for x in reg.vertices]

                width = (np.max(ra_list) - np.min(ra_list)) * units.deg
                height = (np.max(dec_list) - np.min(dec_list)) * units.deg
                sizes = [height, width]

            cutout_list.append((center, sizes, angle))
    else:
        ra_dec = args.center.split(',')

        try:
            ra = float(ra_dec[0])
            dec = float(ra_dec[1])
            center = (ra, dec)
        except ValueError:
            ra = units.Quantity(ra_dec[0])
            dec = units.Quantity(ra_dec[1])
            center = SkyCoord(ra, dec)

        size_list = []
        for size_j in args.size.split(','):
            try:
                size = float(size_j)
            except ValueError:
                size = units.Quantity(size_j)
            size_list.append(size)
        if len(size_list) == 1:
            sizes = (size_list[0], size_list[0])
        else:
            sizes = (size_list[1], size_list[0])

        try:
            angle = float(args.angle) * units.deg
        except ValueError:
            angle = units.Quantity(args.angle)

        cutout_list.append((center, sizes, angle))

    if args.verbose:
        for j, (center, sizes, angle) in enumerate(cutout_list):
            print(f'  - Cutout {j}: {center} {sizes} {angle}')

    target_data_file = args.input_fits[0]

    fits_base_name = os.path.basename(target_data_file)
    fits_base_name = os.path.splitext(fits_base_name)[0]

    data_structure = _get_fits_data_structure(target_data_file)

    if args.verbose:
        print(
            "\n=== IMAGE INFO ===\n"
            f" Name: {fits_base_name}\n"
            f" Type: {data_structure['type']}\n"
            f" Data EXT: {data_structure['data-ext']}\n"
            f" Var EXT: {data_structure['variance-ext']}\n"
            f" DQ EXT: {data_structure['mask-ext']}\n",
            file=sys.stderr
        )
        report_callback = simple_pbar_callback
    else:
        report_callback = None

    with fits.open(target_data_file) as hdul:
        for j, (cut_center, cut_size, cut_angle) in enumerate(cutout_list):
            cutout_name = f"cutout_{fits_base_name}_{j:04}.fits"

            if data_structure['type'] == 'cube':
                with SpectralCube.open(
                    target_data_file,
                    spec_hdu_index=data_structure['data-ext'][0],
                    mode='readonly'
                ) as my_cube:

                    if args.verbose:
                        print(
                            "\nComputing flux cutouts...",
                            file=sys.stderr
                        )
                    flux_cutout = get_cube_cutout(
                        my_cube.spec_hdu.data,
                        center=cut_center,
                        size=cut_size,
                        angle=cut_angle,
                        data_wcs=my_cube.getSpecWCS(),
                        report_callback=report_callback
                    )

                    # Convert specral axis to angtrom units
                    flux_header = updated_wcs_cutout_header(
                        my_cube.spec_hdu.header,
                        flux_cutout['wcs'].to_header()
                    )

                    cutout_hdul = [
                        fits.PrimaryHDU(),
                        fits.ImageHDU(
                            data=flux_cutout['data'],
                            header=flux_header,
                            name=my_cube.spec_hdu.name
                        ),
                    ]

                    if my_cube.var_hdu is not None:
                        if args.verbose:
                            print(
                                "\nComputing variance cutouts...",
                                file=sys.stderr
                            )

                        var_cutout = get_cube_cutout(
                            my_cube.var_hdu.data,
                            center=cut_center,
                            size=cut_size,
                            angle=cut_angle,
                            data_wcs=my_cube.getVarWCS(),
                            report_callback=report_callback
                        )

                        var_header = updated_wcs_cutout_header(
                            my_cube.var_hdu.header,
                            var_cutout['wcs'].to_header()
                        )

                        cutout_hdul.append(
                            fits.ImageHDU(
                                data=var_cutout['data'],
                                header=var_header,
                                name=my_cube.var_hdu.name
                            ),
                        )

                    if my_cube.mask_hdu is not None:
                        if args.verbose:
                            print(
                                "\nComputing data mask cutouts...",
                                file=sys.stderr
                            )

                        mask_cutout = get_cube_cutout(
                            my_cube.mask_hdu.data,
                            center=cut_center,
                            size=cut_size,
                            angle=cut_angle,
                            data_wcs=my_cube.getMaskWCS(),
                            report_callback=report_callback
                        )

                        mask_header = updated_wcs_cutout_header(
                            my_cube.mask_hdu.header,
                            mask_cutout['wcs'].to_header()
                        )

                        cutout_hdul.append(
                            fits.ImageHDU(
                                data=mask_cutout['data'],
                                header=mask_header,
                                name=my_cube.mask_hdu.name
                            ),
                        )

                    if my_cube.wd_hdu is not None:
                        if args.verbose:
                            print(
                                "\nComputing data R matrix cutouts...",
                                file=sys.stderr
                            )

                        wd_cutout = get_cube_cutout(
                            my_cube.wd_hdu.data,
                            center=cut_center,
                            size=cut_size,
                            angle=cut_angle,
                            data_wcs=my_cube.getWdWCS(),
                            report_callback=report_callback
                        )

                        wd_header = updated_wcs_cutout_header(
                            my_cube.wd_hdu.header,
                            wd_cutout['wcs'].to_header()
                        )

                        cutout_hdul.append(
                            fits.ImageHDU(
                                data=wd_cutout['data'],
                                header=wd_header,
                                name=my_cube.wd_hdu.name
                            ),
                        )

                    cutout_hdul = fits.HDUList(cutout_hdul)
                    cutout_hdul.writeto(cutout_name, overwrite=True)
            elif data_structure['type'].endswith('-gray'):
                if data_structure['type'].startswith('image-'):
                    gray_data = hdul[data_structure['data-ext'][0]].data
                else:
                    gray_data = hdul[data_structure['data-ext'][0]].data
                    gray_data = gray_data[..., 0]

                gray_hdu = hdul[data_structure['data-ext'][0]]
                grey_wcs = WCS(gray_hdu.header)

                cutout = get_gray_cutout(
                    gray_data,
                    center=cut_center,
                    size=cut_size,
                    angle=cut_angle,
                    data_wcs=grey_wcs
                )

                gray_header = updated_wcs_cutout_header(
                    gray_hdu.header,
                    cutout['wcs'].to_header()
                )

                cutout_hdul = fits.HDUList([
                    fits.PrimaryHDU(
                        data=cutout['data'],
                        header=gray_header,
                    ),
                ])
                cutout_hdul.writeto(cutout_name, overwrite=True)
            elif data_structure['type'] == 'image-rgb':
                rgb_data = [hdul[k].data for k in data_structure['data-ext']]
                rgb_wcs = [
                    WCS(hdul[k].header) for k in data_structure['data-ext']
                ]

                cutout = get_rgb_cutout(
                    rgb_data,
                    center=cut_center,
                    size=cut_size,
                    angle=cut_angle,
                    data_wcs=rgb_wcs
                )

                header_r = cutout['wcs'][0].to_header()
                header_g = cutout['wcs'][1].to_header()
                header_b = cutout['wcs'][2].to_header()

                cutout_hdul = fits.HDUList([
                    fits.PrimaryHDU(),
                    fits.ImageHDU(
                        data=cutout['data'][0],
                        header=header_r,
                        name='RED',
                    ),
                    fits.ImageHDU(
                        data=cutout['data'][1],
                        header=header_g,
                        name='GREEN',
                    ),
                    fits.ImageHDU(
                        data=cutout['data'][2],
                        header=header_b,
                        name='BLUE',
                    )
                ])
                cutout_hdul.writeto(cutout_name, overwrite=True)
            else:
                print(
                    f"WARNING: not implemente yet [{data_structure['type']}]!",
                    file=sys.stderr
                )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
SPECEX - SPECtra EXtractor.

Extract spectra from spectral data cubes.

Copyright (C) 2022-2023  Maurizio D'Addona <mauritiusdadd@gmail.com>
"""
from __future__ import absolute_import, division, print_function

import os
import sys
import json
import argparse
import logging

from typing import Optional, Any, List, Union

import numpy as np
import multiprocessing

from scipy.signal import savgol_filter
from astropy import wcs
from astropy.table import Table, Column
from astropy.io import fits
from astropy.io.registry import IORegistryError
from astropy.coordinates import SkyCoord
from astropy.time import Time
from astropy import units

import matplotlib.pyplot as plt

from specex.utils import (
    plot_zfit_check, get_log_img, stack, plot_scandata,
    get_spectrum_snr, get_spectrum_snr_emission, parse_regionfile,
    HAS_REGION, HAS_REPROJECT, simple_pbar_callback, do_reproject
)
from specex.sources import get_spectrum
from specex.cube import get_hdu, SpectralCube
from specex.exceptions import exception_handler

from specex.cube import (
    KNOWN_SPEC_EXT_NAMES,
    KNOWN_VARIANCE_EXT_NAMES,
    KNOWN_INVAR_EXT_NAMES,
    KNOWN_MASK_EXT_NAMES,
    KNOWN_RCURVE_EXT_NAMES
)

try:
    from specex.rrspecex import rrspecex, get_templates
except Exception:
    HAS_RR = False
else:
    HAS_RR = True

try:
    from IPython.core import ultratb
    from IPython import embed
except Exception:
    HAS_IPYTHON = False
else:
    HAS_IPYTHON = True

SPECTRA_WEIGHTING_MODES = {
    'none': 'no weighting',
    'whitelight': 'weighting using cube whitelight image',
    'log-whitelight': 'weighting using cube whitelight image in log10 scale',
}

if HAS_REPROJECT:
    SPECTRA_WEIGHTING_MODES['<file url>'] = 'weighting using an external image'
def __argshandler(options: Optional[List[str]] = None) -> Any:
    """
    Parse the arguments given by the user.

    Returns
    -------
    args: Namespace
        A Namespace containing the parsed arguments. For more information see
        the python documentation of the argparse module.
    """
    parser = argparse.ArgumentParser(
        description='Extract spectra from a datacube and optionally compute '
        'the redshift using rrspecex backend.'
    )
    parser.add_argument(
        'input_cube', metavar='SPEC_CUBE', type=str, nargs=1,
        help='The spectral cube in fits format from which spectra will be '
        'extracted.'
    )

    parser.add_argument(
        '--mode', metavar='WORKING_MODE', type=str, default='auto',
        help='Set the working mode of the program. Can be one of the following'
        ' modes: "auto", "kron_ellipse", "kron_circle", "circular_aperture". '
        'When using "kron_ellipse" mode, the input catalog must contain the '
        'column for objetcs kron radius, semi-major and semi-minor axes and '
        'for the rotation angle of the objects. In this mode all the pixels '
        'whitin this ellipse will be added together to compute the spectra. '
        '"kron_circular" mode requires the same parameters of the previous '
        'mode, but a circular aperture is used instead of an ellipse and the '
        'radius of this aperture is the circularized kron radius (i.e. the '
        'kron darius multiplied by the axis ratio q = minor-axis/major-axis. '
        'In circular_aperture mode, no morphological parameters are required '
        'to be in the input catalog and the aperture size is set using the '
        'input parameter --aperture-size. In "auto" mode, the best working '
        'mode is selected based on the parameters available in the input '
        'catalog.'
    )

    parser.add_argument(
        '--catalog', metavar='CATALOG', type=str, default=None,
        help='A catalog containing the position and the (optional) morphology '
        'of the objects of which you want to extract the spectra. The catalog '
        'must contain at least the RA and DEC coordinates and optionally the '
        'dimensions of the isophotal axes and the rotation angle of the '
        ' isophotal ellipse, as well as the kron radius. The way how these '
        'parameter are used depends on the program working mode (see the '
        '--mode section for more information). The catalog format is'
        ' by default compatible by the one generated by Sextractor, so the '
        'column name used to read the aformentioned qauntities are '
        'respectively ALPHA_J2000, DELTA_J2000, A_IMAGE, B_IMAGE, THETA_IMAGE '
        'and KRON_RADIUS. If the catalog has different column names for these '
        'quantities, they can be overridden with a corresponding --key-* '
        'argument (see below). The sky frame of the RA and DEC coordinates can'
        ' be set using --cat-skyframe.'
    )

    parser.add_argument(
        '--regionfile', metavar='REGFILE', type=str, default=None,
        help='A region file containing regions from which extract spectra. '
        'The name of the refion will be used as ID for the spectra.'
    )

    parser.add_argument(
        '--nspectra', metavar='N', type=int, default=0,
        help='Number of spectra to extract (usefull for debugging purposes).'
    )

    parser.add_argument(
        '--sn-threshold', metavar='SN_THRESH', type=float, default=1.5,
        help='Set the signal to Noise ratio threshold: objects with their '
        'spectrum having a SN ratio lower than threshold will be ignored. '
        'If SN_THRESH is set to a value equal or lower than 0, SN filtering '
        'is disabled'
    )

    parser.add_argument(
        '--no-nans', action='store_true', default=False,
        help="This option tells the program to replace eventual NaNs in the"
        "extracted spectrum with linarly interpolated data. In this case the "
        "output fits files contain an additional extension named 'NAN_MASK' "
        "that identifies the previously bad pixels in the spectrum."
    )

    parser.add_argument(
        '--info-hdu', metavar='INFO_HDU', type=int, default=0,
        help='The HDU containing cube metadata. If this argument '
        'Set this to -1 to automatically detect the HDU containing the info. '
        'NOTE that this value is zero indexed (i.e. firts HDU has index 0).'
    )

    parser.add_argument(
        '--spec-hdu', metavar='SPEC_HDU', type=int, default=None,
        help='The HDU containing the spectral data to use. If this argument '
        'Set this to -1 to automatically detect the HDU containing spectra. '
        'NOTE that this value is zero indexed (i.e. second HDU has index 1).'
    )

    parser.add_argument(
        '--var-hdu', metavar='VAR_HDU', type=int, default=None,
        help='The HDU containing the variance of the spectral data. '
        'Set this to -1 if no variance data is present in the cube. '
        'The default value is %(metavar)s=%(default)s.'
        'NOTE that this value is zero indexed (i.e. third HDU has index 2).'
    )

    parser.add_argument(
        '--mask-hdu', metavar='MASK_HDU', type=int, default=None,
        help='The HDU containing the valid pixel mask of the spectral data. '
        'Set this to -1 if no mask is present in the cube. '
        'The default value is %(metavar)s=%(default)s.'
        'NOTE that this value is zero indexed (i.e. fourth HDU has index 3).'
    )

    parser.add_argument(
        '--invert-mask', action='store_true', default=False,
        help='Set whether to use the inverse of the cube mask.'
    )

    parser.add_argument(
        '--aperture-size', metavar='APERTURE', type=str, default='0.8mas',
        help='Set the diameter of a fixed circula aperture used in the '
        '"fixed-aperture" mode. The pixels within this aperture will be '
        'used for spectra extraction. Units can be specified using the names '
        'compatible with astropy.units (eg. 5mas, 1arcsec, etc.).'
        'The default value is %(metavar)s=%(default)s.'
    )

    parser.add_argument(
        '--anulus-size', metavar='R1,R2', type=str, default='None',
        help='Set the inner and outer radii in arcsecs of a fixed anular '
        'aperture used in  the "fixed-aperture" mode. The pixels within this '
        'anulus will be used to compute the background nearby the objects. If '
        'The default value is %(metavar)s=%(default)s.'
    )

    parser.add_argument(
        '--weighting', metavar='MODE', type=str, default='None',
        help='Set the spatial weighting for the spectra. Available modes are:'
        + ';'.join([
            f"{mode_name} for {mode_desc}"
            for mode_name, mode_desc in SPECTRA_WEIGHTING_MODES.items()
        ]) +
        ". Note that in order to use extrenal images as weightmap, the python "
        "package 'reproject' is needed. (Is reproject installed: " +
        "YES" if HAS_REPROJECT else "NO" +
        ")."
    )

    parser.add_argument(
        '--cat-pixelscale', metavar='PIXEL_SCALE', type=str, default='1mas',
        help='Set the scale size in mas of the major and minor axes. Useful if'
        'A and B are given in pixels instead of physical units (like in the '
        'case of catalogs extracted by sextractor). Units can be specified '
        'using the astropy.units names (ie. 5mas, 1arcsec, etc.). If it is '
        'adimensional, then the value is interpreted as mas per pixel. '
        'NOTE that this is not the pixel scale of the cube but it is the size '
        ' of a pixel of the image used to generate the input catalog. '
        'This parameter will be ignored if using a region file as input.'
        'The default value is %(metavar)s=%(default)s.'
    )

    parser.add_argument(
        '--cat-format', metavar='FORMAT', type=str, default=None,
        help='Manually set the format of the input catalog to %(metavar)s. '
        'If not specified the program will try to automatically detect the '
        'format.'
    )

    parser.add_argument(
        '--cutouts-image', metavar='IMAGE', type=str, default=None,
        help='The path of a FITS image (grayscale or single/multiext RGB) from'
        'which to extract cutout of the objects. If this option is not '
        'specified then the cutouts will be extracted directly from the '
        'input cube.'
    )

    parser.add_argument(
        '--cutouts-size', metavar='SIZE', type=float, default=10,
        help='Size of the cutouts of the object in pixel or arcseconds. '
        'The default value is %(metavar)s=%(default)s.'
    )

    parser.add_argument(
        '--key-ra', metavar='RA_KEY', type=str, default='ALPHA_J2000',
        help='Set the name of the column from which to read the RA of the '
        'objects. The default value is %(metavar)s=%(default)s.'
    )

    parser.add_argument(
        '--key-dec', metavar='DEC_KEY', type=str, default='DELTA_J2000',
        help='Set the name of the column from which to read the RA of the '
        'objects. The default value is %(metavar)s=%(default)s.'
    )

    parser.add_argument(
        '--key-a', metavar='A_KEY', type=str, default='A_IMAGE',
        help='Set the name of the column from which to read the size of the'
        'semi-major isophotal axis. The default value is '
        '%(metavar)s=%(default)s.'
    )

    parser.add_argument(
        '--key-b', metavar='B_KEY', type=str, default='B_IMAGE',
        help='Set the name of the column from which to read the size of the'
        'semi-minor isophotal axis. The default value is '
        '%(metavar)s=%(default)s.'
    )

    parser.add_argument(
        '--key-theta', metavar='THETA_KEY', type=str, default='THETA_IMAGE',
        help='Set the name of the column from which to read the rotation angle'
        ' of the isophotal ellipse. The default value is '
        '%(metavar)s=%(default)s.'
    )

    parser.add_argument(
        '--key-kron', metavar='KRON_KEY', type=str, default='KRON_RADIUS',
        help='Set the name of the column from which to read the size of the '
        'Kron radius (in units of A or B). The default value is '
        '%(metavar)s=%(default)s.'
    )

    parser.add_argument(
        '--key-id', metavar='OBJID_KEY', type=str, default=None,
        help='Set the name of the column from which to read the id of the '
        'object to which the spectral data belong. If this argument is not '
        'specified then a suitable column will be selected according to its '
        'name or, if none is found, a new progressive id will be generated. '
        'This id will be included in the fits header.'
    )

    parser.add_argument(
        '--cat-skyframe', metavar='SKY_FRAME', type=str, default='icrs',
        help='Set the sky frame RA and DEC values of the input catalog. This '
        'option is ignored if a region file is uesed for the extraction. '
        'The default value is %(metavar)s=%(default)s.'
    )

    parser.add_argument(
        '--outdir', metavar='DIR', type=str, default=None,
        help='Set the directory where extracted spectra will be outputed. '
        'If this parameter is not specified, then a new directory will be '
        'created based on the name of the input cube.'
    )

    parser.add_argument(
        '--check-images', action='store_true', default=False,
        help='Whether or not to generate check images (cutouts, etc.).'
    )

    parser.add_argument(
        "--checkimg-outdir", type=str, default=None, required=False,
        help='Set the directory where check images are saved (when they are '
        'enabled thorugh the appropriate parameter).'
    )

    if HAS_RR:
        rr_help_note = ''
    else:
        rr_help_note = ' (WARNING: this option is unavailable since redrock' \
                       ' seems not to be installed).'

    parser.add_argument(
        "--zbest", type=str, default=None, required=False,
        help='Whether to find redshift from the spectra. ' + rr_help_note
    )

    parser.add_argument(
        "--priors", type=str, default=None, required=False,
        help='optional redshift prior file. ' + rr_help_note
    )

    parser.add_argument(
        "--nminima", type=int, default=3, required=False,
        help="the number of redshift minima to search. " + rr_help_note
    )

    parser.add_argument(
        "--mp", type=int, default=0, required=False,
        help="if not using MPI, the number of multiprocessing processes to use"
        " (defaults to half of the hardware threads). " + rr_help_note
    )

    parser.add_argument(
        "-t", "--templates", type=str, default=None, required=False,
        help="template file or directory. " + rr_help_note
    )

    parser.add_argument(
        "--debug", action='store_true', default=False,
        help="Enable some debug otupout on plots and on the console."
    )

    args = None
    if options is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(options)

    if args.cutouts_image and not os.path.exists(args.cutouts_image):
        print(f"The file {args.cutouts_image} does not exist!")
        sys.exit(1)

    if args.mp == 0:
        args.mp = int(multiprocessing.cpu_count() / 4)

    return args


# TODO: resample to constant log(lambda)
def get_spplate_fits(cube_header, spec_header, obj_ids, spec_data,
                     var_data=None, and_mask_data=None, or_mask_data=None,
                     wdisp_data=None, sky_data=None):
    """
    Generate a plate fits file similar to the ones used in SDSS.

    Parameters
    ----------
    cube_header : TYPE
        DESCRIPTION.
    spec_header : TYPE
        DESCRIPTION.
    obj_ids : TYPE
        DESCRIPTION.
    spec_data : TYPE
        DESCRIPTION.
    var_data : TYPE, optional
        DESCRIPTION. The default is None.
    and_mask_data : TYPE, optional
        DESCRIPTION. The default is None.
    or_mask_data : TYPE, optional
        DESCRIPTION. The default is None.
    wdisp_data : TYPE, optional
        DESCRIPTION. The default is None.
    sky_data : TYPE, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    hdul : TYPE
        DESCRIPTION.

    """
    spec_hdu = fits.PrimaryHDU()
    ivar_hdu = fits.ImageHDU()
    andmsk_hdu = fits.ImageHDU()
    ormsk_hdu = fits.ImageHDU()
    wdisp_hdu = fits.ImageHDU()
    tbl_hdu = fits.BinTableHDU()
    sky_hdu = fits.ImageHDU()

    if isinstance(spec_data, np.ma.MaskedArray):
        spec_data = spec_data.filled(np.nan)

    if isinstance(var_data, np.ma.MaskedArray):
        var_data = var_data.filled(np.nan)

    spec_hdu.data = spec_data.T
    spec_hdu.header['PLATEID'] = 0

    try:
        spec_hdu.header['MJD'] = cube_header['MJD']
    except KeyError:
        try:
            spec_hdu.header['MJD'] = cube_header['MJD-OBS']
        except KeyError:
            print(
                "WARNING: no MJD found in the cube metadata, using extraction "
                "time instead!",
                file=sys.stderr
            )
            spec_hdu.header['MJD'] = Time.now().mjd

    # See the following link for COEFF0 and COEFF1 definition
    # https://classic.sdss.org/dr7/products/spectra/read_spSpec.html
    try:
        spec_hdu.header["COEFF0"] = spec_header["COEFF0"]
        spec_hdu.header["COEFF1"] = spec_header["COEFF1"]
    except KeyError:
        print(
            "WARING: wavelength binning of the input cube is not compatible "
            "with boss log10 binning data model.\n        Computing the best "
            "values of COEFF0 and COEFF1 for this wavelengt range...",
            file=sys.stderr
        )
        # Compute coeff0 and coeff1 that best approximate wavelength range
        # of the datacube
        crpix = spec_header["CRPIX3"] - 1
        crval = spec_header["CRVAL3"]
        cd1 = spec_header["CD3_3"]

        w_end = spec_data.shape[1] - crpix

        coeff0 = np.log10(crval + cd1*(crpix))
        coeff1 = np.log10(1 + (w_end*cd1)/crval) / w_end

        spec_hdu.header['COEFF0'] = coeff0
        spec_hdu.header['COEFF1'] = coeff1

    spec_hdu.header["CRVAL1"] = spec_header["CRVAL3"]
    spec_hdu.header["CD1_1"] = spec_header["CD3_3"]
    spec_hdu.header["CDELT1"] = spec_header["CD3_3"]
    spec_hdu.header["CRPIX1"] = spec_header["CRPIX3"]
    spec_hdu.header['BUNIT'] = spec_header['BUNIT']
    spec_hdu.header["CUNIT1"] = spec_header["CUNIT3"]
    spec_hdu.header["CTYPE1"] = spec_header["CTYPE3"]

    if var_data is not None:
        ivar_hdu.data = 1 / var_data.T

    if wdisp_data is None:
        wdisp_hdu.data = np.zeros_like(spec_data.T)
    else:
        wdisp_hdu.data = wdisp_data.T

    info_table = Table()
    info_table.add_column(
        Column(data=obj_ids, name='FIBERID', dtype='>i4')
    )

    tbl_hdu.data = info_table.as_array()

    hdul = fits.HDUList([
        spec_hdu, ivar_hdu, andmsk_hdu, ormsk_hdu, wdisp_hdu, tbl_hdu, sky_hdu
    ])

    return hdul


def get_spsingle_fits(main_header, spec_wcs_header, obj_spectrum,
                      spec_hdu_header, obj_spectrum_var=None, no_nans=False):
    """
    Generate a fits containing the spectral data.

    Parameters
    ----------
    main_header : dict
        Dictionary containing the base information of the spectrum extraction.
        The keys of the dictionaries are the names of the fits CARDS that will
        be written, while the values can be either strings or tuples in the
        form of (cadr value, card comment).
    spec_wcs_header : astropy.io.fits.Header object
        Header containing the WCS information of the spectrum.
    obj_spectrum : numpy.ndarray
        The actual spectral data.
    spec_hdu_header : astropy.io.fits.Header object
        Header of the HDU of the original cube containing the spectral data.
    obj_spectrum_var : numpy.ndarray (default=None)
        The variance of the spectral data.
    no_nans : bool, optional
        If true, nan values in the spectrum are replaced by a linear
        interpolation. In this case an additional extension is added to the
        fits file, containing a mask for identifying the previously bax pixels.

    Returns
    -------
    hdul : astropy.io.fits.HDUList object
        The fits HDUL containing the extracted spectra.

    """
    my_time = Time.now()
    my_time.format = 'isot'

    if isinstance(obj_spectrum, np.ma.MaskedArray):
        obj_spectrum = obj_spectrum.filled(np.nan)

    if isinstance(obj_spectrum_var, np.ma.MaskedArray):
        obj_spectrum_var = obj_spectrum_var.filled(np.nan)

    # The primary hdu contains no data but a header with information on
    # how the spectrum was extracted
    main_hdu = fits.PrimaryHDU()
    for key, val in main_header.items():
        try:
            main_hdu.header[key] = val
        except ValueError:
            print(
                f"Invalid value for spec header: {key} = {val}",
                file=sys.stderr
            )

    spec_header = fits.Header()
    spec_header['BUNIT'] = spec_hdu_header['BUNIT']
    spec_header["OBJECT"] = spec_hdu_header["OBJECT"]
    spec_header["CUNIT1"] = spec_hdu_header["CUNIT3"]
    spec_header["CTYPE1"] = spec_hdu_header["CTYPE3"]
    spec_header["OBJECT"] = spec_hdu_header["OBJECT"]
    spec_header["CRPIX1"] = spec_hdu_header["CRPIX3"]
    spec_header["CRVAL1"] = spec_hdu_header["CRVAL3"]
    spec_header["CDELT1"] = spec_hdu_header["CD3_3"]

    nanmask = np.isnan(obj_spectrum)

    if no_nans:
        obj_spectrum[nanmask] = 0
        if obj_spectrum_var is not None:
            obj_spectrum_var[nanmask] = np.nanmax(obj_spectrum_var) * 100.0

    hdu_list = [
        main_hdu,
        fits.ImageHDU(
            data=obj_spectrum,
            header=spec_header,
            name='SPECTRUM'
        )
    ]

    if obj_spectrum_var is not None:
        var_header = fits.Header()
        var_header['BUNIT'] = spec_hdu_header['BUNIT']
        var_header["OBJECT"] = spec_hdu_header["OBJECT"]
        var_header["CUNIT1"] = spec_hdu_header["CUNIT3"]
        var_header["CTYPE1"] = spec_hdu_header["CTYPE3"]
        var_header["OBJECT"] = spec_hdu_header["OBJECT"]
        var_header["CRPIX1"] = spec_hdu_header["CRPIX3"]
        var_header["CRVAL1"] = spec_hdu_header["CRVAL3"]
        var_header["CDELT1"] = spec_hdu_header["CD3_3"]
        var_header["OBJECT"] = spec_hdu_header["OBJECT"]
        hdu_list.append(
            fits.ImageHDU(
                data=obj_spectrum_var,
                header=var_header,
                name='VARIANCE'
            )
        )

    hdu_list.append(
        fits.ImageHDU(
            data=nanmask.astype('uint8'),
            name='NAN_MASK'
        )
    )

    hdul = fits.HDUList(hdu_list)
    return hdul


def load_specex_file(filename, spec_hdu_index=None, var_hdu_index=None,
                     wd_hdu_index=None, mask_hdu_index=None, smoothing=0):
    """
    Load a spectrum file generated by specex (or any compatible file).

    Parameters
    ----------
    filename : str
        The path of the fits file containing the spectrum.
    spec_hdu_index : int or str, optional
        The index or name of the HDU containing the flux.
        The default is None.
    var_hdu_index : int or str, optional
        The index or name of the HDU containing the variance.
        The default is None.
    wd_hdu_index : int or str, optional
        The index or name of the HDU containing the resolution curve.
        The default is None.
    mask_hdu_index : int or str, optional
        The index or name of the HDU containing the flux nan mask.
        The default is None.
    smoothing : int, optional
        If greather than zero, the the spectrum is smoothed with a window of
        this size. The default is 0.

    Raises
    ------
    ValueError
        When flux and variance have different shape.

    Returns
    -------
    dict
        A dict containing the following keys:

        * 'id': the ID of the spectrum,
        * 'main_header': the FITS header of the primary HDU,
        * 'wavelength': the wavelengths of the dispersion grid,
        * 'flux': the fluxes,
        * 'variance': the variance of the fluxes,
        * 'r_curve': the resolution curve,
        * 'nan_mask': a mask indicating invalid fluxes,
        * 'flux_units': the units of the fluxes,
        * 'wavelength_units': the units of the wavelengths

    """
    with fits.open(filename) as hdul:
        valid_id_keys = [
            f"{i}{j}"
            for i in ['', 'OBJ', 'OBJ_', 'TARGET', 'TARGET_']
            for j in ['ID', 'NUMBER', 'UID', 'UUID']
        ]

        spec_id = None
        for hdu in hdul:
            for key in valid_id_keys:
                try:
                    spec_id = hdu.header[key]
                except KeyError:
                    continue
                else:
                    break

        main_header = get_hdu(
            hdul,
            hdu_index=0,
            valid_names=['PRIMARY', 'primary'],
            msg_index_error="WARNING: No Primary HDU",
            exit_on_errors=False
        ).header.copy()
        spec_hdu = get_hdu(
            hdul,
            hdu_index=spec_hdu_index,
            valid_names=KNOWN_SPEC_EXT_NAMES,
            msg_err_notfound="WARNING: No spectrum HDU",
            exit_on_errors=False
        )
        var_hdu = get_hdu(
            hdul,
            hdu_index=var_hdu_index,
            valid_names=KNOWN_VARIANCE_EXT_NAMES,
            msg_err_notfound="WARNING: No variance HDU",
            exit_on_errors=False
        )

        if var_hdu is None:
            ivar_hdu = get_hdu(
                hdul,
                hdu_index=var_hdu_index,
                valid_names=KNOWN_INVAR_EXT_NAMES,
                msg_err_notfound="WARNING: No inverse variance HDU",
                exit_on_errors=False
            )

        mask_hdu = get_hdu(
            hdul,
            hdu_index=mask_hdu_index,
            valid_names=KNOWN_MASK_EXT_NAMES,
            msg_err_notfound="WARNING: No nan mask HDU",
            exit_on_errors=False
        )

        wd_hdu = get_hdu(
            hdul,
            hdu_index=wd_hdu_index,
            valid_names=KNOWN_RCURVE_EXT_NAMES,
            msg_err_notfound="WARNING: No r curve HDU",
            exit_on_errors=False
        )

        smoothing_window_size = 4*smoothing + 1

        if spec_hdu is None:
            flux = None
            flux_units = None
            lam = None
            wavelength_units = None
        else:
            flux = spec_hdu.data.copy()
            spec_wcs = wcs.WCS(spec_hdu.header)
            if smoothing > 0:
                flux = savgol_filter(flux, smoothing_window_size, 3)
            flux = flux.astype('float32')

            pixel = np.arange(len(flux))
            if spec_wcs.has_spectral:
                lam = (
                    spec_wcs.pixel_to_world(pixel)
                    .Angstrom
                    .astype('float32')
                )
            else:
                try:
                    coeff0 = spec_hdu.header["COEFF0"]
                    coeff1 = spec_hdu.header["COEFF1"]
                except KeyError:
                    lam = None
                else:
                    lam = (10**(coeff0 + coeff1*pixel)).astype('float32')

            try:
                flux_units = spec_hdu.header['BUNIT']
            except KeyError:
                flux_units = None

            try:
                wavelength_units = spec_hdu.header['CUNIT1']
            except KeyError:
                wavelength_units = None

        if mask_hdu is None:
            flux_nan_mask = None
        else:
            flux_nan_mask = mask_hdu.data.astype(bool)
            if flux is not None:
                flux_nan_mask |= np.isnan(flux)

        if var_hdu is None:
            if ivar_hdu is None:
                var = None
            else:
                var = 1 / ivar_hdu.data.copy()
                if smoothing > 0:
                    var = savgol_filter(var, smoothing_window_size, 3)
                var = var.astype('float32')
        else:
            var = var_hdu.data.copy()
            if smoothing > 0:
                var = savgol_filter(var, smoothing_window_size, 3)
            var = var.astype('float32')

        if wd_hdu is None:
            wd = None
        else:
            wd = wd_hdu.data.copy()
            if smoothing > 0:
                wd = savgol_filter(wd, smoothing_window_size, 3)
            wd = wd.astype('float32')

        if (
            (flux is not None) and
            (var is not None) and
            flux.shape != var.shape
        ):
            raise ValueError(
                f"'{filename}' - "
                "Spectral data invalid or corruptede: Flux data shape "
                "do not match variance data one!"
            )

    return {
        'id': spec_id,
        'main_header': main_header,
        'wavelength': lam,
        'flux': flux,
        'variance': var,
        'r_curve': wd,
        'nan_mask': flux_nan_mask,
        'flux_units': flux_units,
        'wavelength_units': wavelength_units
    }


def speclist2ez(spec_files, spec_hdu_name='SPECTRUM', var_hdu_name='VARIANCE'):
    """
    Generate a list of spectra that can be imported into pandora EZ.

    Parameters
    ----------
    spec_files : list of str
        List of file paths.
    spec_hdu_name : str, optional
        The name of the HDU containing the spectral data.
        The default is 'SPECTRUM'.
    var_hdu_name : str, optional
        The name of the HDU containing the variance data.
        The default is 'VARIANCE'.

    Returns
    -------
    txt_data : TYPE
        DESCRIPTION.

    """
    txt_data = ""

    for file in spec_files:
        txt_data += f'{file}[{spec_hdu_name}](1) - '
        txt_data += f'{file}[{var_hdu_name}](1) -\n'
    return txt_data


def specex(options=None):
    """
    Run the main program.

    Returns
    -------
    None.

    """
    global HAS_IPYTHON

    # Get user input from argument parsers helper
    args = __argshandler(options)

    if args.debug and HAS_IPYTHON:
        sys.excepthook = exception_handler

    # Working mode
    mode = None

    if args.catalog is not None:
        if args.cat_format is not None:
            sources = Table.read(args.catalog, format=args.cat_format)
        else:
            try:
                sources = Table.read(args.catalog)
            except IORegistryError:
                try:
                    sources = Table.read(args.catalog, format='ascii')
                except IORegistryError:
                    print(
                        "ERROR: cannot determine the format of the input "
                        "catalog. Try to specify it manually using the "
                        "option --cat-format.",
                        file=sys.stderr
                    )
                    sys.exit(1)

        basename_with_ext = os.path.basename(args.catalog)
        pixelscale = units.Quantity(args.cat_pixelscale)
        coord_frame = args.cat_skyframe  # frame of source coordinates
        if str(pixelscale.unit.physical_type) == 'dimensionless':
            pixelscale = pixelscale * units.mas
    elif args.regionfile is not None:
        if not HAS_REGION:
            logging.error(
                "astropy regions package is needed to handle regionfiles!"
            )
            sys.exit(1)

        sources, coord_frame = parse_regionfile(
            args.regionfile,
            key_ra=args.key_ra,
            key_dec=args.key_dec,
            key_id=args.key_id,
            file_format=args.cat_format,
        )
        basename_with_ext = os.path.basename(args.regionfile)
        pixelscale = 1.0 * units.arcsec
        mode = 'regions'
    else:
        logging.error(
            "You must provide at least an input catalog or a region file!",
        )
        sys.exit(1)

    basename = os.path.splitext(basename_with_ext)[0]

    with SpectralCube.open(args.input_cube[0]) as my_cube:
        if args.debug:
            print(f"Flux HDU: {my_cube.spec_hdu.name}")
            if my_cube.var_hdu is not None:
                print(f"Variance HDU: {my_cube.var_hdu.name}")
            if my_cube.mask_hdu is not None:
                print(f"Mask HDU: {my_cube.mask_hdu.name}")

        # Get spectral and celstial WCS from the cube
        celestial_wcs = my_cube.getSpecWCS().celestial
        spectral_wcs = my_cube.getSpecWCS().spectral

        # If weighting with an external image, read that
        valid_wl_modes = ['none', 'whitelight', 'log-whitelight']
        weight_img: Union[None, np.ndarray] = None
        if args.weighting.lower() not in valid_wl_modes:
            if not os.path.isfile(args.weighting):
                logging.error(
                    f"weighting image does not exist '{args.weighting}' "
                )
                sys.exit(1)
            with fits.open(args.weighting) as hdul:
                wimg_hdu = hdul[0]
                wimg_shape = wimg_hdu.data.shape

                if len(wimg_shape) != 2:
                    logging.error(
                        "ERROR: only grayscale images are supported as "
                        "weightmap!"
                    )
                    sys.exit(1)
                elif HAS_REPROJECT:
                    # Reproject the external weight image to the
                    # datacube footprint
                    weight_img, wimg_mask = do_reproject(
                        wimg_hdu, my_cube
                    )

                    weight_img[wimg_mask == 0] = np.nan
                elif wimg_shape != my_cube.spec_hdu.data.shape[1:]:
                    logging.error(
                        "The shape of the external weightmap image does "
                        "not match the datacube and reproject is not "
                        "installed! Use --help for more information."
                    )
                    sys.exit(1)
                else:
                    weight_img = wimg_hdu.data.copy()

        # TODO: we assume here that the distortion across the sky plane is
        #       negligible, but it is very bold assumption.
        #       A more general approach that takes into account also distortion
        #       in function of the position on the sky is needed.
        # Get the pixelscales in arcsec/pixel.
        cube_pixelscale_x, cube_pixelscale_y = [
            units.Quantity(sc_val, u)
            for sc_val, u in zip(
                wcs.utils.proj_plane_pixel_scales(celestial_wcs),
                celestial_wcs.wcs.cunit
            )
        ]

        if args.debug:
            print(
                "Cube X and Y pixelscales are: "
                f"{cube_pixelscale_x.to(units.arcsec)}, "
                f"{cube_pixelscale_y.to(units.arcsec)}",
                file=sys.stderr
            )

        ra_unit = sources[args.key_ra].unit
        if ra_unit is None:
            if args.debug:
                logging.warning(
                    "RA data has no units, assuming values are in degrees!",
                )
            ra_unit = 'deg'

        dec_unit = sources[args.key_dec].unit
        if dec_unit is None:
            if args.debug:
                logging.warning(
                    "DEC data has no units, assuming values are in degrees!",
                )
            dec_unit = 'deg'

        # Get the objects positions in the cube (in pixels)
        obj_sky_coords = SkyCoord(
            sources[args.key_ra], sources[args.key_dec],
            unit=(ra_unit, dec_unit),
            frame=coord_frame
        )

        obj_x, obj_y = wcs.utils.skycoord_to_pixel(
            coords=obj_sky_coords,
            wcs=celestial_wcs
        )

        sources.add_column(obj_x, name='CUBE_X_IMAGE')
        sources.add_column(obj_y, name='CUBE_Y_IMAGE')

        img_height, img_width = my_cube.getSpatialSizePixels()
        yy, xx = np.meshgrid(
            np.arange(img_height),
            np.arange(img_width),
            indexing='ij'
        )

        extracted_data = np.zeros((img_height, img_width))

        # Chech if there is a footprint mask in the cube
        if my_cube.mask_hdu is not None:
            cube_footprint = my_cube.mask_hdu.data.prod(axis=0) == 0
            if args.invert_mask:
                cube_footprint = ~cube_footprint
        else:
            cube_footprint = None

        if args.outdir is None:
            outdir = f'{basename}_spectra'
        else:
            outdir = args.outdir

        if not os.path.isdir(outdir):
            os.mkdir(outdir)

        if mode is None:
            if args.mode == 'auto':
                if (
                        args.key_a in sources.colnames and
                        args.key_b in sources.colnames and
                        args.key_kron in sources.colnames
                ):
                    mode = 'kron_ellipse'
                else:
                    mode = 'circular_aperture'
            else:
                mode = args.mode

        print(f"Extracting spectra with strategy '{mode}'", file=sys.stderr)

        n_objects = len(sources)
        if args.nspectra:
            n_objects = int(args.nspectra)

        # Try to identify the column with object IDs
        if args.key_id is None:
            valid_id_keys = [
                f"{i}{j}"
                for i in ['', 'OBJ', 'OBJ_', 'TARGET', 'TARGET_']
                for j in ['ID', 'NUMBER', 'UID', 'UUID']
            ]

            for key in valid_id_keys:
                if key.lower() in sources.colnames:
                    key_id = key.lower()
                    break
                elif key.upper() in sources.colnames:
                    key_id = key.upper()
                    break
            else:
                key_id = None
        else:
            key_id = args.key_id

        out_specfiles = []
        specex_apertures = {}  # these values are in physical units (arcsecs)
        specex_anuli = {}  # these values are in physical units (arcsecs)
        source_ids = []

        valid_sources_mask = np.zeros(len(sources), dtype=bool)

        # Check wheter to use anuli to compute background
        anulus_str = args.anulus_size.lower().strip().split(',')

        anulus_r_in = None
        anulus_r_out = None
        if anulus_str and len(anulus_str) == 2:
            try:
                anulus_r1 = float(anulus_str[0])
                anulus_r2 = float(anulus_str[1])
            except ValueError:
                pass
            else:
                anulus_r_in = min(anulus_r1, anulus_r2)
                anulus_r_out = max(anulus_r1, anulus_r2)

        if args.debug:
            print(
                f"Read {len(sources[:n_objects])} sources from "
                "input catalog/regionfile...",
                file=sys.stderr
            )

        for i, source in enumerate(sources[:n_objects]):
            simple_pbar_callback(i + 1, n_objects)

            if key_id is not None:
                obj_id = source[key_id]
            else:
                obj_id = f"{i:06}"

            obj_ra = source[args.key_ra]
            obj_dec = source[args.key_dec]
            xx_tr = xx - source['CUBE_X_IMAGE']
            yy_tr = yy - source['CUBE_Y_IMAGE']

            anulus_mask = None
            if mode == 'regions':
                anulus_mask = None
                pix_reg = source['region'].to_pixel(celestial_wcs)
                reg_mask = pix_reg.to_mask()
                mask = reg_mask.to_image((img_height, img_width)) > 0
                specex_apertures[obj_id] = (
                    1 * pixelscale,
                    1 * pixelscale,
                    0 * units.rad
                )

            elif mode == 'kron_ellipse':
                ang = np.deg2rad(source[args.key_theta])

                x_over_a = xx_tr*np.cos(ang) + yy_tr*np.sin(ang)
                x_over_a /= source[args.key_a]

                y_over_b = xx_tr*np.sin(ang) - yy_tr*np.cos(ang)
                y_over_b /= source[args.key_b]

                specex_apertures[obj_id] = (
                    source[args.key_a] * pixelscale,
                    source[args.key_b] * pixelscale,
                    source[args.key_theta] * units.rad
                )

                mask = (x_over_a**2 + y_over_b**2) < (1.0/source[args.key_kron])

            elif mode == 'kron_circular':
                kron_circular = source[args.key_kron] * source[args.key_b]
                kron_circular /= source[args.key_a]

                specex_apertures[obj_id] = (
                    kron_circular * pixelscale,
                    kron_circular * pixelscale,
                    0 * units.rad
                )

                mask = (xx_tr**2 + yy_tr**2) < (kron_circular ** 2)

            elif mode == 'circular_aperture':
                aperture_size = units.Quantity(args.aperture_size)
                aperture_radius = aperture_size / 2
                if str(aperture_size.unit.physical_type) == 'dimensionless':
                    aperture_radius = aperture_radius * units.mas

                try:
                    specex_apertures[obj_id] = (
                        aperture_radius,
                        aperture_radius,
                        0 * units.rad
                    )
                except TypeError:
                    logging.warning(
                        f"object {obj_id} has an invalid ID, skipping..."
                    )
                    continue

                xx_scaled = xx_tr * cube_pixelscale_x
                yy_scaled = yy_tr * cube_pixelscale_y

                mask = (xx_scaled**2 + yy_scaled**2) < (aperture_radius)**2

                # TODO: fix here
                """
                if anulus_r_in and anulus_r_out:
                    specex_anuli[obj_id] = (
                        anulus_r_in,
                        anulus_r_out
                    )

                    anulus_mask_out = rad_coords <= anulus_r_out / cube_pixelscale
                    anulus_mask_in = rad_coords > anulus_r_in / cube_pixelscale

                    anulus_mask = anulus_mask_in & anulus_mask_out
                else:
                    specex_anuli[obj_id] = None
                    anulus_mask = None
                """

            if cube_footprint is not None:
                mask &= cube_footprint

            if np.sum(mask) == 0:
                logging.warning(
                    f"object {obj_id} has no footprint, skipping..."
                )
                continue

            obj_extraction_mask = np.zeros_like(mask, dtype='float32')
            if args.weighting.lower() == 'none':
                weights = None
                obj_extraction_mask[mask] = 1.0
            else:
                if args.weighting.lower() == 'whitelight':
                    # summing selected spaxels along spectral axis to get a
                    # whitelight weightmap
                    w_arr = np.ma.array(
                        my_cube.spec_hdu.data[:, mask],
                        mask=~np.isfinite(my_cube.spec_hdu.data[:, mask]),
                        dtype='float32'
                    )
                    weights = np.ma.mean(w_arr, axis=0)
                    weights -= np.ma.min(weights)
                elif args.weighting.lower() == 'log-whitelight':
                    w_arr = np.ma.array(
                        my_cube.spec_hdu.data[:, mask],
                        mask=~np.isfinite(my_cube.spec_hdu.data[:, mask]),
                        dtype='float32'
                    )
                    weights = np.ma.mean(w_arr, axis=0)
                    weights -= np.ma.min(weights)
                    weights = np.ma.log10(1 + weights)
                elif weight_img is not None:
                    weights = weight_img[mask]
                    weights -= np.ma.min(weights)

                obj_extraction_mask[mask] = weights

            var_spaxels = None
            if my_cube.var_hdu is not None:
                var_spaxels = my_cube.var_hdu.data[:, mask]

            obj_spectrum, obj_spectrum_var = get_spectrum(
                flux_spaxels=my_cube.spec_hdu.data[:, mask],
                var_spaxels=var_spaxels,
                weights=weights
            )

            obj_spectrum = obj_spectrum.filled(np.nan)
            obj_spectrum_var = obj_spectrum_var.filled(np.nan)

            if np.sum(
                np.isnan(obj_spectrum_var[~np.isnan(obj_spectrum)])
            ) != 0:
                logging.warning(
                    f"\nWARNING: object {obj_id}  is weird! Skipping...\n"
                )
                continue

            if np.sum(~np.isnan(obj_spectrum)) < 100:
                logging.warning(
                    f"WARNING: object {obj_id} has invalid data..."
                )

            if anulus_mask is not None:
                # TODO: fix here!
                # Here we suppose that background vaies very slowly within
                # object aperture. In this case we can copute the mean
                # background and subtract it total contribution within the
                # aperture.
                bkg_spectrum = np.median(
                    my_cube.spec_hdu.data[:, mask], axis=0
                )
                smoothed_bkg_spectrum = savgol_filter(bkg_spectrum, 51, 11)
                obj_spectrum -= smoothed_bkg_spectrum*np.sum(mask)

            if np.sum(~np.isnan(obj_spectrum)) == 0:
                if args.debug:
                    logging.warning(
                        f"WARNING: object {obj_id} has no spectral data, "
                        "skipping..."
                    )
                continue

            sn_spec = get_spectrum_snr(obj_spectrum, obj_spectrum_var)
            sn_em = get_spectrum_snr_emission(obj_spectrum, obj_spectrum_var)

            if (
                    (args.sn_threshold > 0) and
                    (max(sn_spec, sn_em) < args.sn_threshold)
               ):

                if args.debug:
                    logging.warning(
                        f"WARNING: object {obj_id} has SNR={sn_spec:.2f} < "
                        f"{args.sn_threshold}, skipping..."
                    )
                continue

            my_time = Time.now()
            my_time.format = 'isot'

            outname = f"spec_{obj_id}.fits"

            apertures_info = [
                x.to_string() for x in specex_apertures[obj_id]
            ]

            main_header = {
                'CUBE': (
                    basename_with_ext, "Spectral cube used for extraction"
                ),
                'RA': (
                    obj_ra, "Ra of the center of the object"
                ),
                'DEC': (
                    obj_dec, "Dec of the center of the object"
                ),
                'FRAME': (
                    coord_frame, "Ra and Dec sky frame"
                ),
                'NPIX': (
                    np.sum(mask), 'Number of pixels used for this spectra'
                ),
                'HISTORY': f'Extracted on {str(my_time)}',
                'ID': obj_id,
                'SN': (
                    sn_spec, "SNR of the total spectrum"
                ),
                'SN_EMISS': (
                    sn_em, "SNR due to emission lines only"
                ),
                'EXT_MODE': (
                    mode, "Spex extraction mode"
                ),
                'EXT_APER': (
                    json.dumps(apertures_info),
                    "Apertures used in specex"
                )
            }

            # Copy the spectral part of the WCS into the new FITS
            spec_wcs_header = spectral_wcs.to_header()

            hdul = get_spsingle_fits(
                main_header, spec_wcs_header,
                obj_spectrum, my_cube.spec_hdu.header,
                obj_spectrum_var, no_nans=args.no_nans
            )

            out_file_name = os.path.join(outdir, outname)
            hdul.writeto(
                out_file_name,
                overwrite=True
            )

            # If we arrived here, then the object has a vail spectrum and can
            # be saved. Let's also mark the object as valid and store its ID
            out_specfiles.append(os.path.realpath(out_file_name))
            source_ids.append(obj_id)
            valid_sources_mask[i] = True

            # Add also the extracted pixels to the extraction map
            extracted_data += obj_extraction_mask
            if anulus_mask is not None:
                extracted_data -= anulus_mask

        # Discard all invalid sources
        sources = sources[valid_sources_mask]
        source_ids = np.array(source_ids)

        with open("spec_list_pandora_ez.txt", 'w') as f:
            f.write(speclist2ez(out_specfiles))

        if args.debug:
            print(
                f"\nExtracted {len(sources)} valid sources\n",
                file=sys.stderr
            )

        if args.checkimg_outdir is not None:
            check_images_outdir = args.checkimg_outdir
        else:
            check_images_outdir = os.path.join(outdir, 'checkimages')

        if not os.path.isdir(check_images_outdir):
            os.mkdir(check_images_outdir)

        # NOTE: for some weird reason redrock hangs if any figure has been
        #       created using matplotlib. Do all the plottings only after
        #       redrock has finished!
        if args.zbest:
            rrspecex_options = [
                '--zbest', args.zbest,
                '--checkimg-outdir', check_images_outdir
            ]

            if args.priors is not None:
                rrspecex_options += ['--priors', args.priors]

            if args.nminima is not None:
                rrspecex_options += ['--nminima', f'{args.nminima:d}']

            if args.mp is not None:
                rrspecex_options += ['--mp', f'{args.mp:d}']

            if args.templates is not None:
                rrspecex_options += ['--templates', f'{args.templates}']

            rrspecex_options += out_specfiles
            targets, zfit, scandata = rrspecex(options=rrspecex_options)

            if args.debug or not args.cutouts_image:
                stacked_cube = stack(my_cube.spec_hdu.data)

            if args.debug:
                fig = plt.figure(figsize=(10, 10))
                ax = plt.subplot(projection=celestial_wcs)
                logimg, cvmin, cvmax = get_log_img(stacked_cube)
                ax.imshow(
                    logimg,
                    origin='lower',
                    vmin=cvmin,
                    vmax=cvmax,
                    cmap='jet'
                )
                fig.savefig("stacked_cube.png")
                plt.tight_layout()
                plt.close(fig)

            if args.debug:
                plot_templates = get_templates(templates=args.templates)
            else:
                plot_templates = None

            print("", file=sys.stderr)

            # Creating directory to save Chi-2 plots
            if args.debug:
                x2_plots_out_dir = os.path.join(
                    check_images_outdir,
                    "X2_plots"
                )
                if not os.path.isdir(x2_plots_out_dir):
                    os.mkdir(x2_plots_out_dir)

            for i, target in enumerate(targets):
                logging.info("Generating previews...")
                simple_pbar_callback(i, n_objects-1)

                fig, axs = plot_zfit_check(
                    target,
                    zfit,
                    plot_template=plot_templates,
                    wavelengt_units=my_cube.spec_hdu.header['CUNIT3'],
                    flux_units=my_cube.spec_hdu.header['BUNIT'],
                )

                figname = f'spectrum_{target.spec_id}.png'
                figname = os.path.join(check_images_outdir, figname)
                fig.savefig(figname, dpi=150)
                plt.close(fig)

                if args.debug:
                    figname = f'scandata_{target.spec_id}.png'
                    figname = os.path.join(x2_plots_out_dir, figname)
                    fig, axs = plot_scandata(target, scandata)
                    fig.savefig(figname, dpi=150)
                    plt.close(fig)

        specex_map = fits.PrimaryHDU(
            data=extracted_data,
            header=celestial_wcs.to_header()
        )
        specex_map.writeto(
            f"{my_cube.getBaseName()}_extmap.fits",
            overwrite=True
        )

        if args.check_images:
            fig, ax = plt.subplots(1, 1, figsize=(10, 10))
            ax.imshow(extracted_data, origin='lower')
            fig.savefig("sext_extraction_map.png", dpi=150)
            plt.close(fig)


if __name__ == '__main__':
    specex()



"""
SPEX - SPectra EXtractor.

Emission and absorption line dictionary

Copyright (C) 2022-2024  Maurizio D'Addona <mauritiusdadd@gmail.com>

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from typing import Optional, Union
import numpy as np
from scipy.stats import median_abs_deviation
from scipy.signal import savgol_filter
from scipy.signal import find_peaks_cwt

from astropy.modeling import Fittable1DModel, Parameter

# Some important lines with corresponding wavelenghts in Angstrom
RESTFRAME_LINES = [
    (10320, '[SII]', 'E'),
    (8863.0, 'TiO', 'A'),
    (8430.0, 'TiO', 'A'),
    (8195.0, 'NaI', 'A'),
    (8183.0, 'NaI', 'A'),
    (7590.0, 'TiO', 'A'),
    (7065.2, 'HeI', 'AE'),
    (6725.0, '[SII]', 'E'),
    (6562.8, 'Halpha', 'AEB'),
    (6159.0, 'TiO', 'A'),
    (5892.5, 'NaD', 'A'),
    (5603.0, 'TiO', 'A'),
    (5269.0, 'Ca,Fe', 'A'),
    (5175.4, 'MgI', 'A'),
    (5006.8, '[OIII]', 'E'),
    (4958.9, '[OIII]', 'E'),
    (4861.3, 'Hbeta', 'AEB'),
    (4340.4, 'Hgamma', 'AE'),
    (4304.4, 'Gband', 'A'),
    (4216.0, 'CN', 'A'),
    (4101.7, 'Hdelta', 'AE'),
    (4000.0, 'Balmer_Break', 'Break'),
    (4072.0, '[SII]', 'E'),
    (3968.5, 'CaII_H', 'A'),
    (3933.7, 'CaII_K', 'A'),
    (3889.1, 'Hksi,CN(H8)', 'AE'),
    (3869.0, '[NeIII]', 'E'),
    (3797.9, 'Hteta', 'AE'),
    (3770.6, 'H11', 'AE'),
    (3727.5, '[OII]', 'E'),
    (3581.0, 'FeI', 'A'),
    (3425.8, '[NeV]', 'E'),
    (3345.9, '[NeV]', 'E'),
    (2964.0, 'FeII_bump', 'E'),
    (2799.0, 'MgII', 'AEB'),
    (2626.0, 'FeII', 'E'),
    (2600.0, 'FeII', 'A'),
    (2586.7, 'FeII', 'A'),
    (2382.0, 'FeII', 'A'),
    (2374.0, 'FeII', 'A'),
    (2344.2, 'FeII', 'A'),
    (2260.0, 'FeII', 'A'),
    (2142.0, '[NII]', 'E'),
    (1909.0, '[CIII]', 'EB'),
    (1856.0, 'AlIII', 'A'),
    (1670.8, 'AlII', 'A'),
    (1666.1497, 'OIII]', 'E'),
    (1640.0, 'HeII', 'AE'),
    (1608.5, 'FeII', 'A'),
    (1660.8092, 'OIII]', 'E'),
    (1549.0, 'CIV', 'AEB'),
    (1526.7, 'SiII', 'A'),
    (1397.0, 'SiIV+OIV', 'AEB'),
    (1334.5, 'CII', 'AE'),
    (1303.0, 'OI', 'AE'),
    (1260.4, 'SiII', 'A'),
    (1240.0, 'NV', 'AE'),
    (1215.7, 'LyA', 'AEB'),
    (1033.0, 'OVI', 'AE'),
    (1025.6, 'LyB', 'AE'),
    (972.5, 'LyG', 'AE'),
]


def _normal(x, mu, sigma):
    return np.exp(-((x - mu)**2)/(2*sigma)) / (sigma * np.sqrt(2 * np.pi))


class Emission1D(Fittable1DModel):
    """Simple model for a flat spectrum with emission lines."""

    redshift = Parameter()

    def __init__(self, lines_identifications, sigma=3, redshift=0, **kwargs):
        self.line_candidates = lines_identifications
        self.sigma = sigma
        super().__init__(redshift=redshift, **kwargs)

    def evaluate(self, lam, redshift):
        """
        Evaluate the model.

        Parameters
        ----------
        lam : np.ndarray
            Array of wavelength.
        redshift : float, optional
            The redshift of the spectrum.
            The default value is 0.

        Returns
        -------
        result : np.ndarray
            The model values.

        """
        result = np.zeros_like(lam)
        for candidate in self.line_candidates:
            mu = candidate[1] / (1 + redshift)
            result += _normal(lam, mu, self.sigma) * candidate[3]
        return result


def get_lines(name=None, line_type=None, wrange=None, z=0):
    """
    Return line data according to the given line name and types.

    Parameters
    ----------
    name : str or None, optional
        The name of the line (eg. CaII_H or FeI, etc...). If None, the lines
        are selected only by type. If both name and type are None, all lines
        are returned.
    line_type : str or None, optional
        Type of the line, can be 'A' (absorption), 'E' (emission) 'B' (Broad).
        If None, then all the line types are returned.
        The default is None.
    wrange : tuple/list/np.ndarray of floats or None, optional
        The wavelength ragne in which lines should be. If None, no selection
        according to the line wavelenght is made.
        The default is None.
    z : float, optional
        The redshit of the lines. The default value is 0.

    Returns
    -------
    selected_lines : list
        List of line data. Each element of the list is a 3-tuple in the form
        (wavelenght in Angstrom, Line name, Line type).

    """
    if name is None:
        selected_lines = RESTFRAME_LINES[:]
    else:
        selected_lines = [
            line
            for line in RESTFRAME_LINES
            if name.lower() == line[1].lower()
        ]
    if line_type is not None:
        selected_lines = [
            line
            for line in selected_lines
            if line_type.lower() in line[2].lower()
        ]

    selected_lines = [
        ((1 + z) * line[0], line[1], line[2])
        for line in selected_lines
    ]

    if wrange is not None:
        w_min = np.nanmin(wrange)
        w_max = np.nanmax(wrange)
        selected_lines = [
            line
            for line in selected_lines
            if w_min <= line[0] <= w_max
        ]

    return selected_lines


def get_spectrum_lines(
        wavelengths: np.ndarray,
        flux: np.ndarray,
        var: Optional[np.ndarray] = None,
        sigma_threshold: Optional[float] = 10.0,
        smoothing_window: Optional[int] = 51,
        smoothing_order: Optional[int] = 1,
        glob_smoothing: Optional[int] = 3):
    """
    Identify the position of clear emission or absorption lines.

    Parameters
    ----------
    wavelengths : np.ndarray
        The wavelengs corresponding to each flux value.
    flux : numpy.ndarray
        The spectrum itself.
    var : numpy.ndarray, optional
        The variance of the spectrum itself.
        The default value is None.
    sigma_threshold : float, optional
        The threshold to use for line identification.
        The default value is 5.0
    smoothing_window : int, optional
        Parameter to be passed to the smoothing function.
        The default value is 51.
    smoothing_order : int, optional
        Parameter to be passed to the smoothing function.
        The default value is 11.
    glob_smoothing : int, optional
        Parameter used for line identification.
        The default value is 3

    Returns
    -------
    identifications : list
        A list of tuple. Each tuple ha the form of (k, w, l, h) and contains
        the index k for the wavelenght w of the line, the approximate max width
        l of the line and a height h of the line. Note that l and h are not
        actual phisical quantities and should be used with caution when
        comparing to other values from a different spectrum.
    """
    if np.isnan(flux).all():
        return np.nan
    else:
        flux = savgol_filter(flux.copy(), glob_smoothing, 1)
        flux = np.ma.array(flux, mask=np.isnan(flux))

    if var is not None:
        var = np.ma.array(var.copy(), mask=np.isnan(var))
    else:
        var = 1.0

    smoothed_spec = savgol_filter(flux, smoothing_window, smoothing_order)
    smoothed_spec = np.ma.array(smoothed_spec, mask=np.isnan(smoothed_spec))

    # Subtract the smoothed spectrum to the spectrum itself to get a
    # crude estimation of the noise, then square it and divide for the variance
    # and then go back with a square root
    norm_noise = ((flux - smoothed_spec)**2) / var
    norm_noise = np.ma.sqrt(norm_noise)

    # Get the median value of the noise. The median is more robust against the
    # presence of lines with respect to the mean
    noise_median = np.ma.median(norm_noise)

    # Get the NMAD of the noise. We assume here that the noise has a
    # unimodal distribution (eg. gaussian like), and this is a good assumption
    # if the noise is due only to the random fluctuations
    noise_nmad = median_abs_deviation(norm_noise, scale='normal')

    norm_noise_deb = np.abs(norm_noise - noise_median)

    # Get the possible lines
    outlier = norm_noise_deb >= (sigma_threshold * noise_nmad)

    # Delete identification with lenght 1 (almost all are fake)
    for k, v in enumerate(outlier):
        if (k == 0) or (k == len(outlier)-1):
            continue
        if v and ((outlier[k-1] == 0) and (outlier[k+1] == 0)):
            outlier[k] = 0

    # Merge almost contiguos identifications
    for k in range(1, glob_smoothing+1):
        outlier[k:] += outlier[:-k]
        outlier[:-k] += outlier[k:]

    # Get position, width and height of the identifications
    identifications = []
    c_start = None
    c_wstart = None
    c_end = None
    for k, v in enumerate(outlier):
        if v:
            if c_start is None:
                c_start = k
                c_wstart = wavelengths[k]
                c_end = None
        elif c_start is not None:
            if c_end is None:
                c_end = k
                c_wh = np.ma.max(norm_noise_deb[c_start: c_end])
                c_wh /= noise_nmad
                c_max_pos = np.ma.argmax(norm_noise_deb[c_start: c_end])
                c_pos_idx = c_start + c_max_pos
                c_wpos = wavelengths[c_pos_idx]
                c_wlen = wavelengths[k] - c_wstart
                identifications.append((c_pos_idx, c_wpos, c_wlen, c_wh))
                c_start = None

    # Sortin by height
    identifications.sort(key=lambda a: a[3])
    return identifications


def get_redshift_from_lines(identifications: Union[tuple, list, np.array],
                            z_max: Optional[float] = 6,
                            z_min: Optional[float] = 0,
                            z_points: Optional[Union[float, None]] = None,
                            tol: Optional[float] = None):
    """
    Get the redshift of a set of line identifications.

    Parameters
    ----------
    identifications : Union[tuple, list, np.array]
        A list of identification generated by get_spectrum_lines().
    z_max : Optional[float], optional
        The maximum redshift. The default is 6.
    z_min : Optional[float], optional
        The minimum redshift. The default is 0.
    z_points : Optional[Union[float, None]], optional
        Number of redshift values between z_min and z_max to test.
        If None, then z_points = 1000*(z_max - z_min).
        The default is None.
    tol : Optional[float], optional
        The tolerance. If None, it is computed automatically.
        The default is None.

    Returns
    -------
    z_values
        Best estimations of the redshift sorted from the most probable to the
        least probable.
    z_probs
        Pseudo-probabilities of the redshift estimations (the higer the better)
    """
    if len(identifications) < 2:
        return None

    if z_points is None:
        z_points = 1000 * (z_max - z_min)

    if tol is None:
        tol = np.mean([x[2] for x in identifications])

    mymodel = Emission1D(identifications, tol, redshift=0)

    z_values = np.linspace(z_min, z_max, z_points)
    prob_values = np.zeros_like(z_values)
    for j, z in enumerate(z_values):
        rest_lines_lam = [
            x[0] for x in get_lines(z=z)
        ]

        prob_values[j] = np.sum(mymodel(rest_lines_lam))

    peak_indices = find_peaks_cwt(prob_values, 1)
    z_values_p = z_values[peak_indices]
    z_prob_p = prob_values[peak_indices]

    z_prob_p_sorted_ind = np.argsort(z_prob_p)[::-1]
    z_values_p = z_values_p[z_prob_p_sorted_ind]
    z_prob_p = z_prob_p[z_prob_p_sorted_ind]

    mean_prob = np.median(z_prob_p)
    std_prob = np.std(z_prob_p)

    plausible_mask = z_prob_p >= mean_prob + std_prob

    return (z_values_p[plausible_mask], z_prob_p[plausible_mask])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
SPECEX - SPECtra EXtractor.

This module provides functions to calculate the zeropoints of a FITS image.

For more information take a look at the following links:

 - https://www.stsci.edu/hst/wfpc2/Wfpc2_dhb/wfpc2_ch52.html
 - https://www.stsci.edu/hst/instrumentation/acs/data-analysis/zeropoints

Copyright (C) 2022-2023  Maurizio D'Addona <mauritiusdadd@gmail.com>
"""
import sys
import argparse
import numpy as np
from astropy.io import fits
from astropy import units


def get_zeropoint_info(filename, ext=0):
    """
    Compute zero-point information from a fits image.

    For more information about zeropoints take a look at the following links:
    https://www.stsci.edu/hst/wfpc2/Wfpc2_dhb/wfpc2_ch52.html
    https://www.stsci.edu/hst/instrumentation/acs/data-analysis/zeropoints

    Parameters
    ----------
    filename : str
        The path of the fits file image.

    ext : int, optional
        The FITS extension containing the image data. The default value is 0.

    Returns
    -------
    zpt_dict : ditct
        A dicttionary containing the zero-point information.
        The dictionary contains the following items:

            - exp_time : float
                The exposure time

            - phot_f_lam : float
                The value of PHOTFLAM. This is the flux of a source with
                constant flux per unit wavelength (in erg s-1 cm-2 A-1) which
                produces a count rate of 1 count per second.

            - phot_p_lam : float
                the pivot wavelength (in Angstrom)

            - zero_point : sloat
                the zero point value

            - zero_point_p : float
                the zero point value plus 2.5 times log10 of the exposure time

            - zero_point_m : float
                the zero point value minus 2.5 times log10 of the exposure time

            - counts_to_flux : float
                Quantity to convert counts to flux units
    """
    hdr = fits.getheader(filename, ext=ext)
    phot_f_lam = hdr['PHOTFLAM']
    phot_p_lam = hdr['PHOTPLAM']
    exp_time = hdr['EXPTIME']

    try:
        science_units = hdr['BUNIT']
    except KeyError:
        science_units = None

    zpt = -2.5 * np.log10(phot_f_lam) - 21.10
    zpt += -5 * np.log10(phot_p_lam) + 18.6921

    acs_zpt_pexp = zpt + 2.5 * np.log10(exp_time)
    acs_zpt_mexp = zpt - 2.5 * np.log10(exp_time)

    if science_units is not None and science_units.lower().endswith('/s'):
        counts_to_flux = phot_f_lam
    else:
        counts_to_flux = phot_f_lam/exp_time

    zpt_dict = {
        "exp_time": exp_time,
        "phot_f_lam": phot_f_lam,
        "phot_p_lam": (phot_p_lam / 10) * units.nm,
        "zero_point": zpt,
        "zero_point_p": acs_zpt_pexp,
        "zero_point_m": acs_zpt_mexp,
        'counts_to_flux': counts_to_flux
    }
    return zpt_dict


def print_zeropoint_info(filename, ext=0):
    """
    Print the zero-points information of a FITS image.

    Parameters
    ----------
    filename : str
        The path of a FITS file image.
    ext : int, optional
        The FITS extension containing the image data. The default value is 0.

    Returns
    -------
    None.

    """
    zpt_dict = get_zeropoint_info(filename)
    print(f"\n{filename}")
    s = "Exp time: {exp_time}\n"
    s += "Pivot wavelenght: {phot_p_lam:.0f}\n"
    s += "Zero point: {zero_point}\n"
    s += "Zero point (+m): {zero_point_p}\n"
    s += "Zero point (-m): {zero_point_m}\n"
    print(s.format(**zpt_dict))


def main(options=None):
    """
    Run the main program of this module.

    Returns
    -------
    None.

    """
    parser = argparse.ArgumentParser(
        description='Print zero-point ifnormatio of FITS images.'
    )
    parser.add_argument(
        'inp_files', metavar='FITS_IMAGES', type=str, nargs='+',
        help='One or more than one fits file for which you want to view the'
        'zero-point information'
    )
    parser.add_argument(
        '--ext', metavar='EXT', type=int, default=0,
        help='The extension containing the image data.'
     )

    if options is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(options)

    for fname in args.inp_files:
        try:
            print_zeropoint_info(fname)
        except Exception:
            print(
                f"ERROR: Cannot read zeropoint info for file {fname}",
                file=sys.stderr
            )
            continue


if __name__ == '__main__':
    main()
